"""Exec module for managing Amazon API Gateway Deployments."""
import copy
from collections import OrderedDict
from typing import Any
from typing import Dict
from typing import List


def convert_raw_deployment_to_present(
    hub,
    raw_resource: Dict[str, Any],
    rest_api_id: str,
    resource_id: str,
    idem_resource_name: str = None,
) -> Dict[str, Any]:
    resource_parameters = OrderedDict(
        {
            "description": "description",
        }
    )

    resource_translated = {
        "resource_id": resource_id,
        "rest_api_id": rest_api_id,
    }

    if idem_resource_name:
        resource_translated["name"] = idem_resource_name

    for parameter_raw, parameter_present in resource_parameters.items():
        if raw_resource.get(parameter_raw) is not None:
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)

    return resource_translated


async def update(
    hub,
    ctx,
    old_state: Dict[str, Any],
    updatable_parameters: Dict[str, Any],
):
    """

    Updates an API Deployment Resource. Updating is only supported for replacing descriptions.

    Args:
        old_state(dict):
            Previous state of the resource.

        updatable_parameters(dict):
            Parameters from SLS File.

    Returns:
        {"result": True|False, "comment": A message List, "ret": None|Dict}

    """

    result = dict(comment=(), result=True, ret=None)
    patch_ops = []

    new_state = copy.deepcopy(old_state)
    updated = False

    for key, value in updatable_parameters.items():
        if old_state.get(key) and value is not None and value != old_state.get(key):
            patch_ops.append(
                {
                    "op": "replace",
                    "path": f"/{key}",
                    "value": str(value)
                    if not isinstance(value, List)
                    else ",".join(value),
                }
            )
            updated = True

    if ctx.get("test", False):
        if updated:
            result["comment"] = hub.tool.aws.comment_utils.would_update_comment(
                resource_type="aws.apigateway.deployment",
                name=old_state["name"],
            )
            result["ret"] = new_state
        else:
            result["comment"] = hub.tool.aws.comment_utils.already_exists_comment(
                resource_type="aws.apigateway.deployment",
                name=old_state["name"],
            )
        return result

    if updated:
        update_ret = await hub.exec.boto3.client.apigateway.update_deployment(
            ctx,
            restApiId=old_state["rest_api_id"],
            deploymentId=old_state["resource_id"],
            patchOperations=patch_ops,
        )
        if not update_ret["result"]:
            result["result"] = False
            result["comment"] = update_ret["comment"]
            return result
        else:
            result["comment"] = hub.tool.aws.comment_utils.update_comment(
                resource_type="aws.apigateway.deployment",
                name=old_state["name"],
            )
            result["ret"] = new_state
    else:
        result["comment"] = hub.tool.aws.comment_utils.already_exists_comment(
            resource_type="aws.apigateway.deployment", name=old_state["name"]
        )

    return result



"""Read & extract service metadata and its available operations"""
import copy
import re

import boto3
from botocore.exceptions import UnknownServiceError


def parse_resource_and_operations(
    hub, service_name: str, session: "boto3.session.Session"
):
    """
    Get resource and their available operations for client initialized for a given service

    @returns
        Mapping of resource to its methods and corresponding boto3 operation_name.
        {
            "resource": {
                { "method" : "operation_name" }
            }
        }
    """
    operations = {}

    try:
        client = session.client(service_name=service_name, region_name="us-west-2")
    except UnknownServiceError as ex:
        hub.log.error(f"{ex.__class__.__name__}: {ex}")
        return {}

    # Attempt 1: Lookup resources from service model
    # Supported services: cloudformation, cloudwatch, dynamodb, ec2, glacier, iam, opsworks, s3, sns, sqs
    try:
        service = session.resource(service_name=service_name, region_name="us-west-2")
        service_resources = [
            hub.tool.format.case.snake(r.name)
            for r in service.meta.resource_model.subresources
        ]
        for op in sorted(client.meta.method_to_api_mapping):
            try:
                if "tag" in op:
                    # No need to parse tag functions here
                    continue

                _, resource = op.split("_", maxsplit=1)
                singular = None
                if re.match(rf"\w+[^aoius]s$", resource):
                    singular = hub.tool.format.inflect.singular(resource)
                if resource in service_resources or singular in service_resources:
                    if resource not in operations:
                        operations[resource] = []

                    if op not in operations[resource]:
                        operations[resource].append(op)

            except ValueError:
                continue
    except Exception:
        # It is okay if we cannot get resources from service model, we fall back to default operations lookup
        pass

    # Attempt 2: Now let's look up using all available operations
    for op in sorted(client.meta.method_to_api_mapping):
        try:
            if "tag" in op:
                # No need to parse tag functions here
                continue

            _, resource = op.split("_", maxsplit=1)

            if resource.endswith("apis"):
                resource = resource[:-1]

            # Check for singularity of the resource
            if re.match(rf"\w+[^aoius]s$", resource):
                resource = hub.tool.format.inflect.singular(resource)

            # No need for empty resource to go further
            if resource == "":
                continue

            if resource not in operations:
                operations[resource] = []

            if op not in operations[resource]:
                operations[resource].append(op)

        except ValueError:
            hub.log.error("Failure in extracting operation metadata")

    # Get all potential resource operations (GET/CREATE/UPDATE/DELETE/LIST and TOOL module operations)
    return hub.pop_create.aws.service.get_resource_and_operations(operations)


def get_resource_and_operations(hub, aws_operations: dict):
    resource_and_operations = {}
    for r, ops in aws_operations.items():
        resource_and_operations[r] = {}
        # This will hold all tools operations for a resource
        resource_and_operations[r]["tools"] = []
        for op in sorted(ops):
            # Determine if it is a reserved function or a utility method.
            func_type = hub.pop_create.aws.possible_functions.get_possible_func_type(
                resource_name=r, operation=op
            )

            if func_type == "tools":
                resource_and_operations[r]["tools"].append(op)
            else:
                resource_and_operations[r][func_type] = op

        if (
            "get" not in resource_and_operations[r]
            and "list" in resource_and_operations[r]
        ):
            # majority cases, this is going to be okay. There could be some exceptions.
            resource_and_operations[r]["get"] = resource_and_operations[r]["list"]

        if (
            "update" not in resource_and_operations[r]
            and "create" in resource_and_operations[r]
            and resource_and_operations[r]["create"].startswith("put")
        ):
            # Many "put" operations do support both create and update.
            resource_and_operations[r]["update"] = resource_and_operations[r]["create"]

    for r, ops in resource_and_operations.items():
        # 1. If there is no create operation, most likely, all operations are utility methods on the resource.
        # 2. If there is only single reserved operation other than tools functions, they go into tools module.
        non_utility_ops = copy.copy(ops)
        del non_utility_ops["tools"]
        if "create" not in ops or len(non_utility_ops) < 2:
            other_operations = {
                operation for key, operation in ops.items() if key not in "tools"
            }
            if "tools" in resource_and_operations[r]:
                resource_and_operations[r]["tools"].extend(list(other_operations))
            else:
                resource_and_operations[r] = {"tools": other_operations}

            resource_and_operations[r] = {k: v for k, v in ops.items() if k == "tools"}

    # format :
    # "resource_name": {
    #     "create": ".....",
    #     "delete": "....",
    #     "list": ".....",
    #     "update": ".....",
    #     "get": ".....",
    #     "tools": [
    #         ".....",
    #         ".....",
    #     ]
    # },
    return resource_and_operations


def parse_docstring(hub, session: "boto3.session.Session", service_name: str):
    """
    Get service description
    """
    client = session.client(service_name=service_name, region_name="us-west-2")
    plugin_docstring = hub.tool.format.html.parse(client._service_model.documentation)
    return "\n".join(hub.tool.format.wrap.wrap(plugin_docstring, width=120))


def parse_service_tag_methods(
    hub, session: "boto3.session.Session", aws_service_name: str
):
    """
    Parses service tag method definitions. There is usually a common method at service level which can be used for
    tagging. Sometimes it is single method for update and sometimes there are separate add/remove methods.
    Capture them all here.
    """
    tag_methods = dict()
    try:
        client = session.client(service_name=aws_service_name, region_name="us-west-2")
        for op in client.meta.method_to_api_mapping:
            if re.match("(add|create|put|tag).*_(resource|tags|tagging)", op):
                tag_methods["tag_resource"] = hub.pop_create.aws.function.parse(
                    client=client,
                    aws_service_name=aws_service_name,
                    resource_name=None,
                    func_name=op,
                )
            elif re.match("(remove|delete|untag).*_(resource|tags|tagging)", op):
                tag_methods["untag_resource"] = hub.pop_create.aws.function.parse(
                    client=client,
                    aws_service_name=aws_service_name,
                    resource_name=None,
                    func_name=op,
                )
            elif re.match("(list|get|describe).*_(tags|tagging)", op):
                tag_methods["list_tags"] = hub.pop_create.aws.function.parse(
                    client=client,
                    aws_service_name=aws_service_name,
                    resource_name=None,
                    func_name=op,
                )
            elif re.match("(change).*_(tags)", op):
                tag_methods["update_tags"] = hub.pop_create.aws.function.parse(
                    client=client,
                    aws_service_name=aws_service_name,
                    resource_name=None,
                    func_name=op,
                )
            else:
                continue
    except Exception as err:
        hub.log.error(
            f"Error when generating tag action definitions for {aws_service_name}: {err.__class__.__name__}: {err}"
        )

    return tag_methods



from typing import Any
from typing import Dict

try:
    import pgpy

    HAS_LIBS = (True,)
except ImportError as e:
    HAS_LIBS = False, str(e)


def __virtual__(hub):
    return HAS_LIBS


__func_alias__ = {"list_": "list"}


async def list_(
    hub, ctx, user_name: str, access_key_id: str = None, status: str = None
) -> Dict[str, Any]:
    """
    Returns a list of access keys for a user_name, optionally filtering by status and access_key_id.

    Args:
        user_name (str):
            The user_name to list keys for.

        access_key_id (str, Optional):
            match exact access key to return

        status (str, Optional):
            match "Active" or "Inactive" keys; "Active" means the key is valid for API calls.

    Returns:
        dict[str, Any]:
            Returns a list of access keys for the user

    Examples:
        Calling this exec module function from the cli

        .. code-block:: bash

            idem exec aws.iam.access_key.list user_name="a_user"

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: aws.iam.access_key.list
                - kwargs:
                    user_name: a_user
    """
    # we manually filter status, so we must validate it
    if status not in [None, "Active", "Inactive"]:
        return {
            "result": False,
            "ret": None,
            "comment": (
                f"Could not list access keys: '{status}' isn't a valid status",
            ),
        }

    ret = await hub.exec.boto3.client.iam.list_access_keys(ctx, UserName=user_name)

    if not ret["result"]:
        return {
            "result": False,
            "ret": None,
            "comment": (f"Could not list access keys: {ret['comment']}",),
        }

    access_keys = []
    for access_key_raw in ret["ret"]["AccessKeyMetadata"]:
        tmp_access_key = (
            hub.tool.aws.iam.conversion_utils.convert_raw_access_key_to_camel_case(
                access_key_raw
            )
        )
        if access_key_id and access_key_id != tmp_access_key.get("access_key_id"):
            continue
        if status and status != tmp_access_key.get("status"):
            continue
        idem_resource_name = f"{tmp_access_key.get('user_name', '')}-{tmp_access_key.get('access_key_id', '')}"
        access_key = hub.tool.aws.iam.conversion_utils.convert_access_key_to_present(
            tmp_access_key, idem_resource_name=idem_resource_name
        )
        access_keys.append(access_key)

    result = {"result": True, "ret": access_keys, "comment": ()}
    if not len(access_keys):
        if status and access_key_id:
            result["comment"] += (
                f"List for {user_name} was successful, but no key matched {access_key_id} and {status}",
            )
        elif access_key_id:
            result["comment"] += (
                f"List for {user_name} was successful, but no key matched {access_key_id}",
            )
        elif status:
            result["comment"] += (
                f"List for {user_name} was successful, but no key matched {status}",
            )
        else:
            result["comment"] += (
                f"List for {user_name} was successful, but there were no keys.",
            )

    return result


async def update(
    hub, ctx, user_name: str, access_key_id: str, status: str
) -> Dict[str, Any]:
    """
    Updates the status of an existing access key. Does not execute if ctx["test"] is True.

    Args:
        user_name (str):
            Exact user that owns the key

        access_key_id (str):
            Exact access key to update

        status (str):
            "Active" or "Inactive", meaning it is valid for API calls

    Returns:
        dict[str, Any]:
            Status of the update

    Examples:
        Calling this exec module function from the cli

        .. code-block:: bash

            idem exec aws.iam.access_key.update user_name="a_user" access_key_id="ABCDEF0123456789ABCD" status="Active"

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: aws.iam.access_key.update
                - kwargs:
                    user_name: a_user
                    access_key_id: ABCDEF0123456789ABCD
                    status: Active
    """
    result = {"result": True, "ret": None, "comment": ()}
    # this should never happen and would result in unexpected behavior
    if (not user_name) or (not access_key_id) or (not status):
        result["result"] = False
        result["comment"] = (
            "user_name, access_key_id and status all need to be supplied",
        )
        return result

    ret = await hub.exec.boto3.client.iam.update_access_key(
        ctx, AccessKeyId=access_key_id, Status=status, UserName=user_name
    )

    if not ret["result"]:
        result["result"] = False
        result["comment"] += (
            f"Could not update access key {access_key_id} for {user_name}",
        )
        result["comment"] += ret["comment"]
    else:
        result["comment"] += (
            f"Access Key {access_key_id} for {user_name} set to '{status}'",
        )
    return result


async def delete(hub, ctx, user_name: str, access_key_id: str) -> Dict[str, Any]:
    """
    Deletes an existing access key, does not execute if ctx["test"] is set.

    "result" is True if delete succeeded or no such key or user exists.
    "result" is False if there is an error deleting the key

    Args:
        user_name (str):
            Exact user owning the key

        access_key_id (str):
            Exact access key to delete

    Returns:
        dict[str, Any]:
            Status of the deletion

    Examples:
        Calling this exec module function from the cli

        .. code-block:: bash

            idem exec aws.iam.access_key.delete user_name="a_user" access_key_id="ABCDEF0123456789ABCD"

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: aws.iam.access_key.delete
                - kwargs:
                    user_name: a_user
                    access_key_id: ABCDEF0123456789ABCD
    """
    result = {"result": True, "ret": None, "comment": ()}
    # this should never happen and would result in unexpected behavior
    assert user_name and access_key_id

    ret = await hub.exec.boto3.client.iam.delete_access_key(
        ctx, AccessKeyId=access_key_id, UserName=user_name
    )

    if not ret["result"]:
        if "NoSuchEntityException" in str(ret["comment"]):
            result["comment"] = (
                f"Access key {access_key_id} for {user_name} does not exist",
            )
        else:
            result["result"] = False
            result["comment"] = (
                f"Could not delete access key '{access_key_id}'",
            ) + ret["comment"]
    else:
        result["comment"] = (f"Access key '{access_key_id}' deleted",)
    return result


async def create(hub, ctx, user_name: str, pgp_key: str = None) -> Dict[str, Any]:
    """
    Creates a new access key for the specified user.

    Args:
        user_name (str):
            Exact user the key will be created for.
        pgp_key (str, Optional):
            Base 64 encoded public pgp key with which we will encrypt the returned secret_access_key.

    Returns:
        dict[str, Any]:
            Access key ID, Secret access key, status

    Examples:
        Calling this exec module function from the cli

        .. code-block:: bash

            idem exec aws.iam.access_key.create user_name="a_user" pgp_key="-----BEGIN PGP PUBLIC KEY BLOCK-----0123456789ABCDEFxxxETC-----END PGP PUBLIC KEY BLOCK-----"

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: aws.iam.access_key.create
                - kwargs:
                    user_name: a_user
                    pgp_key:
                        -----BEGIN PGP PUBLIC KEY BLOCK-----
                        0123456789ABCDEFxxxETC
                        -----END PGP PUBLIC KEY BLOCK-----
    """

    result = {"result": True, "ret": None, "comment": ()}

    ret = await hub.exec.boto3.client.iam.create_access_key(ctx, UserName=user_name)

    if not ret["result"]:
        result["result"] = False
        result["comment"] += (f"Could not create access key: {ret['comment']}",)
    else:
        access_key = (
            hub.tool.aws.iam.conversion_utils.convert_raw_access_key_to_camel_case(
                ret["ret"]["AccessKey"]
            )
        )
        if pgp_key:
            decoded_pgp_key = hub.tool.aws.b64.decode(pgp_key)
            loaded_pgp_key, _ = pgpy.PGPKey.from_blob(decoded_pgp_key)
            secret_access_key = pgpy.PGPMessage.new(
                access_key["secret_access_key"],
                compression=pgpy.constants.CompressionAlgorithm.Uncompressed,
            )
            encrypted_secret_access_key = loaded_pgp_key.encrypt(secret_access_key)
            access_key["secret_access_key"] = hub.tool.aws.b64.encode(
                encrypted_secret_access_key
            )

        result["ret"] = access_key
    return result



"""States module for managing ACM certificates."""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List


__contracts__ = ["resource"]
__reconcile_wait__ = {"static": {"wait_in_seconds": 60}}

acceptors = [
    {
        "matcher": "error",
        "expected": "ResourceNotFoundException",
        "state": "retry",
        "argument": "Error.Code",
    },
    {
        "matcher": "path",
        "expected": "ISSUED",
        "state": "success",
        "argument": "Certificate.Status",
    },
    {
        "matcher": "path",
        "expected": "PENDING_VALIDATION",
        "state": "success",
        "argument": "Certificate.Status",
    },
    {
        "matcher": "path",
        "expected": "FAILED",
        "state": "success",
        "argument": "Certificate.Status",
    },
    {
        "matcher": "path",
        "expected": "VALIDATION_TIMED_OUT",
        "state": "success",
        "argument": "Certificate.Status",
    },
    {
        "matcher": "path",
        "expected": "INACTIVE",
        "state": "success",
        "argument": "Certificate.Status",
    },
    {
        "matcher": "path",
        "expected": "EXPIRED",
        "state": "success",
        "argument": "Certificate.Status",
    },
    {
        "matcher": "path",
        "expected": "REVOKED",
        "state": "success",
        "argument": "Certificate.Status",
    },
]


async def present(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    certificate: str = None,
    private_key: str = None,
    certificate_chain: str = None,
    domain_name: str = None,
    validation_method: str = None,
    subject_alternative_names: List[str] = None,
    idempotency_token: str = None,
    domain_validation_options: List[
        make_dataclass(
            "DomainValidationOption",
            [
                ("domain_name", str),
                ("validation_domain", str),
                ("validation_method", str, field(default=None)),
                (
                    "resource_record",
                    make_dataclass(
                        "ResourceRecord",
                        [
                            ("Name", str, field(default=None)),
                            ("Type", str, field(default=None)),
                            ("Value", str, field(default=None)),
                        ],
                    ),
                    field(default=None),
                ),
                ("validation_status", str, field(default=None)),
            ],
        )
    ] = None,
    options: make_dataclass(
        "CertificateOptions",
        [("CertificateTransparencyLoggingPreference", str, field(default=None))],
    ) = None,
    certificate_authority_arn: str = None,
    timeout: make_dataclass(
        "Timeout",
        [
            (
                "create",
                make_dataclass(
                    "CreateTimeout",
                    [
                        ("delay", int, field(default=None)),
                        ("max_attempts", int, field(default=None)),
                    ],
                ),
                field(default=None),
            )
        ],
    ) = None,
    tags: Dict[str, Any]
    or List[
        make_dataclass(
            "Tag",
            [("Key", str, field(default=None)), ("Value", str, field(default=None))],
        )
    ] = None,
) -> Dict[str, Any]:
    """Manage SSL/TLS certificates for your AWS-based websites and applications.

    Imports a certificate into AWS Certificate Manager (ACM) to use with services that are integrated with ACM.

    Note that integrated services allow only certificate types and keys they support to be associated with their
    resources. Further, their support differs depending on whether the certificate is imported into IAM or into ACM.
    Requests an ACM certificate for use with other AWS services. To request an ACM certificate, you must specify a
    fully qualified domain name (FQDN) in the DomainName parameter. You can also specify additional FQDNs in the
    SubjectAlternativeNames parameter.

    If you are requesting a private certificate, domain validation is not required. If you are requesting a public
    certificate, each domain name that you specify must be validated to verify that you own or control the domain.
    You can use DNS validation or email validation.

    Args:
        name(str):
            An Idem name of the resource.

        resource_id(str, Optional):
            The Amazon Resource Name (ARN) of certificate to identify the resource.

        certificate(bytes, Optional):
            The certificate to import.

        private_key(bytes, Optional):
            The private key that matches the public key in the certificate.

        certificate_chain(bytes, Optional):
            The PEM encoded certificate chain.

        domain_name(str, Optional):
            Fully qualified domain name (FQDN), such as www.example.com, that you want to secure with an ACM certificate.

        validation_method(str, Optional):
            The method you want to use if you are requesting a public certificate to validate that you own or control domain.

        subject_alternative_names(list[str], Optional):
            Additional FQDNs to be included in the Subject Alternative Name extension of the ACM certificate.

        idempotency_token(str, Optional):
            Customer chosen string that can be used to distinguish between calls to RequestCertificate.

        domain_validation_options(list[dict[str, Any]], Optional):
            The domain name that you want ACM to use to send you emails so that you can validate domain ownership.

            Defaults to None.

            * domain_name (str):
                A fully qualified domain name (FQDN) in the certificate request.

            * validation_domain (str):
                The domain name that you want ACM to use to send you validation emails. This domain name is the
                suffix of the email addresses that you want ACM to use. This must be the same as the DomainName
                value or a superdomain of the DomainName value.

                For example, if you request a certificate for testing.example.com, you can specify example.com for this value.
                In that case, ACM sends domain validation emails to the following five addresses:
                admin@example.com, administrator@example.com, hostmaster@example.com, postmaster@example.com, webmaster@example.com

            * validation_method (str):
                Specifies the domain validation method.

            * resource_record (dict[str, Any]):
                Contains the CNAME record that you add to your DNS database for domain validation.

                * Name (str):
                    The name of the DNS record to create in your domain. This is supplied by ACM.

                * Type (str):
                    The type of DNS record. Currently this can be CNAME.

                * Value (str):
                    The value of the CNAME record to add to your DNS database. This is supplied by ACM.

            * validation_status (str):
                The validation status of the domain name.
                This can be one of the following values:

                    * PENDING_VALIDATION
                    * SUCCESS
                    * FAILED

        options(dict[str, Any], Optional):
            Currently, you can use this parameter to specify whether to add the certificate to a certificate
            transparency log. Certificate transparency makes it possible to detect SSL/TLS certificates that
            have been mistakenly or maliciously issued. Certificates that have not been logged typically
            produce an error message in a browser. For more information, see Opting Out of Certificate
            Transparency Logging. Defaults to None.

            * CertificateTransparencyLoggingPreference (str, Optional):
                You can opt out of certificate transparency logging by specifying the DISABLED option. Opt in by specifying ENABLED.

        certificate_authority_arn(str, Optional):
            The Amazon Resource Name (ARN) of the private certificate authority (CA) that will be used to issue the certificate.

        tags(dict[str, str], Optional):
            The collection of tags associated with the certificate. Defaults to None.

            * Key (str):
                The key of the tag.

            * Value (str):
                The value of the tag.

        timeout(dict[str, Any], Optional):
            Timeout configuration for request/import AWS Certificate.

            * create(dict[str, int]:
                Timeout configuration for request/importing AWS Certificate.

                * delay (int, Optional):
                    The amount of time in seconds to wait between attempts.

                * max_attempts (int, Optional):
                    Customized timeout configuration containing delay and max attempts.

    Request Syntax:
        .. code-block:: sls

            [certificate-resource-id]:
              aws.acm.certificate_manager.present:
                - domain_name: 'string'
                - validation_method: 'string'
                - subject_alternative_names:
                  - 'string'
                - idempotency_token: 'string'
                - domain_validation_options:
                  - domain_name: 'string'
                    validation_domain: 'string'
                    validation_method: 'string'
                    resource_record:
                      Name: 'string'
                      Value: 'string'
                      Type: 'string'
                    validation_status: PENDING_VALIDATION|SUCCESS|FAILED
                - options:
                  CertificateTransparencyLoggingPreference: ENABLED|DISABLED
                - certificate_authority_arn: 'string'
                - tags:
                  - 'string': 'string'
                - timeout:
                  create:
                    delay: 'int'
                    max_attempts: 'int'

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

          # Request a certificate
          [certificate-resource-id]:
            aws.acm.certificate_manager.present:
              - domain_name: www.example.com
              - validation_method: DNS
              - subject_alternative_names:
                  - www.example.net
              - idempotency_token: ExampleIdempotancyToken
              - domain_validation_options:
                - domain_name: testing.example.com
                  validation_domain: example.com
              - options:
                  CertificateTransparencyLoggingPreference: DISABLED
              - certificate_authority_arn: arn:aws:acm-pca:region:account:certificate-authority/12345678-1234-1234-1234-123456789012
              - tags:
                - class: test

          # Import a certificate
          [certificate-resource-id]:
            aws.acm.certificate_manager.present:
              - resource_id: arn:aws:acm-pca:region:account:certificate-authority/12345678-1234-1234-1234-123456789012
              - certificate_arn: arn:aws:acm-pca:region:account:certificate-authority/12345678-1234-1234-1234-123456789012
              - certificate: example_certificate
              - private_key: -----BEGIN RSA PRIVATE KEY-----
                  MIIEpQIBAAKCAQEA3Tz2mr7SZiAMfQyuvBjM9Oi..Z1BjP5CE/Wm/Rr500P
                  RK+Lh9x5eJPo5CAZ3/ANBE0sTK0ZsDGMak2m1g7..3VHqIxFTz0Ta1d+NAj
                  -----END RSA PRIVATE KEY-----
              - certificate_chain: example_certificate_chain
              - tags:
                - class: test
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None

    if isinstance(tags, List):
        tags = hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags)

    if resource_id:
        before = await hub.exec.aws.acm.certificate_manager.get(
            ctx,
            name=name,
            resource_id=resource_id,
        )

        if not before["result"] or not before["ret"]:
            result["comment"] = before["comment"]
            result["result"] = False
            return result

        result["old_state"] = copy.deepcopy(before["ret"])
        result["new_state"] = copy.deepcopy(result["old_state"])

        are_tags_updated = False
        is_reimported = False

        try:
            plan_state = copy.deepcopy(result["old_state"])

            # Reimport certificate if certificate arn,private key and certificate body is given as input
            # In Idem we cannot check if these properties are changed because they are not retrievable from other APIs
            # We cannot provide tags when reimporting a certificate.
            if private_key and certificate:
                re_imported = await hub.exec.boto3.client.acm.import_certificate(
                    ctx=ctx,
                    CertificateArn=resource_id,
                    Certificate=certificate,
                    PrivateKey=private_key,
                    CertificateChain=certificate_chain,
                )
                result["result"] = re_imported["result"]
                if not result["result"]:
                    result["comment"] = re_imported["comment"]
                    return result
                is_reimported = re_imported["result"]
                result["comment"] = result["comment"] + (
                    f"Re-Imported aws.acm.certificate_manager '{name}'",
                )
                if ctx.get("test", False) and is_reimported:
                    plan_state["private_key"] = private_key
                    plan_state["certificate"] = certificate
                    if certificate_chain:
                        plan_state["certificate_chain"] = certificate_chain
                    result["comment"] = result["comment"] + (
                        f"Would reimport certificate for aws.acm.certificate_manager {name}",
                    )
            if tags is not None and tags != result["old_state"].get("tags"):
                # Update tags
                update_tags_ret = await hub.tool.aws.acm.tag.update_tags(
                    ctx=ctx,
                    resource_id=resource_id,
                    old_tags=result["old_state"].get("tags"),
                    new_tags=tags,
                )
                if not update_tags_ret["result"]:
                    result["comment"] = result["comment"] + update_tags_ret["comment"]
                    result["result"] = False
                    return result
                are_tags_updated = bool(update_tags_ret["ret"])
                if ctx.get("test", False) and are_tags_updated:
                    plan_state["tags"] = update_tags_ret["ret"]
                    result["comment"] = result["comment"] + (
                        f"Would update tags for aws.acm.certificate_manager {name}",
                    )
                elif are_tags_updated:
                    result["comment"] = result["comment"] + (
                        f"Updated tags for aws.acm.certificate_manager '{name}'",
                    )
            if not are_tags_updated and not is_reimported:
                result["comment"] = result["comment"] + (
                    f"aws.acm.certificate_manager '{name}' has no property that needs to be updated",
                )
        except Exception as e:
            result["comment"] = result["comment"] + (str(e),)
            result["result"] = False
    else:
        # If private key is provided as an input then certificate needs to be imported
        # To import new certificate, CertificateArn should not be provided
        if private_key:
            try:
                if ctx.get("test", False):
                    result[
                        "new_state"
                    ] = hub.tool.aws.test_state_utils.generate_test_state(
                        enforced_state={},
                        desired_state={
                            "name": name,
                            "certificate": certificate,
                            "private_key": private_key,
                            "certificate_chain": certificate_chain,
                            "tags": tags,
                        },
                    )
                    result["comment"] = result["comment"] + (
                        f"Would import aws.acm.certificate_manager {name}",
                    )
                    return result
                imported = await hub.exec.boto3.client.acm.import_certificate(
                    ctx=ctx,
                    Certificate=certificate,
                    PrivateKey=private_key,
                    CertificateChain=certificate_chain,
                    Tags=hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags)
                    if tags
                    else None,
                )
                result["result"] = imported["result"]
                if not result["result"]:
                    result["comment"] = result["comment"] + imported["comment"]
                    return result

                resource_id = imported["ret"]["CertificateArn"]

                # Custom waiter for import
                # When the certificate is just created, there is a delay of several seconds before we can retrieve information about it.
                waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
                    default_delay=10,
                    default_max_attempts=6,
                    timeout_config=timeout.get("create") if timeout else None,
                )
                certificate_waiter = hub.tool.boto3.custom_waiter.waiter_wrapper(
                    name="CertificateCreated",
                    operation="DescribeCertificate",
                    argument=["Error.Code", "Certificate.Status"],
                    acceptors=acceptors,
                    client=await hub.tool.boto3.client.get_client(ctx, "acm"),
                )
                await hub.tool.boto3.client.wait(
                    ctx,
                    "acm",
                    "CertificateCreated",
                    certificate_waiter,
                    CertificateArn=resource_id,
                    WaiterConfig=waiter_config,
                )

                result["comment"] = result["comment"] + (
                    f"Imported aws.acm.certificate_manager '{name}'",
                )

            except Exception as e:
                result["comment"] = result["comment"] + (str(e),)
                result["result"] = False
        # If domain name is provided as an input then certificate needs to be requested/created
        elif domain_name:
            try:
                if ctx.get("test", False):
                    result[
                        "new_state"
                    ] = hub.tool.aws.test_state_utils.generate_test_state(
                        enforced_state={},
                        desired_state={
                            "name": name,
                            "domain_name": domain_name,
                            "validation_method": validation_method,
                            "subject_alternative_names": subject_alternative_names,
                            "idempotency_token": idempotency_token,
                            "domain_validation_options": domain_validation_options,
                            "options": options,
                            "certificate_authority_arn": certificate_authority_arn,
                            "tags": tags,
                        },
                    )
                    result["comment"] = result[
                        "comment"
                    ] + hub.tool.aws.comment_utils.would_create_comment(
                        resource_type="aws.acm.certificate_manager", name=name
                    )
                    return result

                requested = await hub.exec.boto3.client.acm.request_certificate(
                    ctx=ctx,
                    DomainName=domain_name,
                    ValidationMethod=validation_method,
                    SubjectAlternativeNames=subject_alternative_names,
                    IdempotencyToken=idempotency_token,
                    DomainValidationOptions=domain_validation_options,
                    Options=options,
                    CertificateAuthorityArn=certificate_authority_arn,
                    Tags=hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags)
                    if tags
                    else None,
                )
                result["result"] = requested["result"]
                if not result["result"]:
                    result["comment"] = result["comment"] + requested["comment"]
                    return result
                resource_id = requested["ret"]["CertificateArn"]

                # Custom waiter for request
                # When the certificate is just created, there is a delay of several seconds before we can retrieve information about it.
                waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
                    default_delay=10,
                    default_max_attempts=6,
                    timeout_config=timeout.get("create") if timeout else None,
                )
                certificate_waiter = hub.tool.boto3.custom_waiter.waiter_wrapper(
                    name="CertificateCreated",
                    operation="DescribeCertificate",
                    argument=["Error.Code", "Certificate.Status"],
                    acceptors=acceptors,
                    client=await hub.tool.boto3.client.get_client(ctx, "acm"),
                )
                await hub.tool.boto3.client.wait(
                    ctx,
                    "acm",
                    "CertificateCreated",
                    certificate_waiter,
                    CertificateArn=resource_id,
                    WaiterConfig=waiter_config,
                    delay=30,
                )

                result["comment"] = result[
                    "comment"
                ] + hub.tool.aws.comment_utils.create_comment(
                    resource_type="aws.acm.certificate_manager", name=name
                )

            except Exception as e:
                result["comment"] = result["comment"] + (str(e),)
                result["result"] = False
                return result
        # If neither of private key and domain name is provided
        else:
            hub.log.debug(
                "AWS Certificate must be imported (private_key) or created (domain_name)"
            )
            result["comment"] = (
                "AWS Certificate must be imported (private_key) or created (domain_name)",
            )
            result["result"] = False
            return result
    try:
        if ctx.get("test", False):
            result["new_state"] = plan_state
        elif not (before and before["result"]) or are_tags_updated or is_reimported:
            after = await hub.exec.aws.acm.certificate_manager.get(
                ctx,
                name=name,
                resource_id=resource_id,
            )
            if not after["result"]:
                result["result"] = False
                result["comment"] = after["comment"]
                return result
            result["new_state"] = copy.deepcopy(after["ret"])
        else:
            result["new_state"] = copy.deepcopy(result["old_state"])

    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False
    return result


async def absent(hub, ctx, name: str, resource_id: str = None) -> Dict[str, Any]:
    """Deletes a certificate and its associated private key.

    If this action succeeds, the certificate no longer appears in the list that can be displayed by calling the
    ListCertificates action or be retrieved by calling the GetCertificate action.
    The certificate will not be available for use by Amazon Web Services services integrated with ACM.

    Args:
        name(str):
            An Idem name of the resource.

        resource_id(str, Optional):
            The Amazon Resource Name (ARN) of certificate to identify the resource.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            resource_is_absent:
              aws.acm.certificate.absent:
                - name: value
                - resource_id: value
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.acm.certificate_manager", name=name
        )
        return result
    before = await hub.exec.aws.acm.certificate_manager.get(
        ctx=ctx, name=name, resource_id=resource_id
    )
    if not before["result"]:
        result["comment"] = before["comment"]
        result["result"] = False
        return result
    if not before["ret"]:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.acm.certificate_manager", name=name
        )
    elif ctx.get("test", False):
        result["old_state"] = before["ret"]
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.acm.certificate_manager", name=name
        )
        return result
    else:
        result["old_state"] = before["ret"]
        ret = await hub.exec.boto3.client.acm.delete_certificate(
            ctx, CertificateArn=resource_id
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.acm.certificate_manager", name=name
        )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Returns detailed metadata about ACM certificates.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: bash

            $ idem describe aws.acm.certificate
    """
    result = {}
    certificate_list_ret = await hub.exec.boto3.client.acm.list_certificates(
        ctx,
        Includes={
            "keyTypes": [
                "RSA_1024",
                "RSA_2048",
                "RSA_3072",
                "RSA_4096",
                "EC_prime256v1",
                "EC_secp384r1",
                "EC_secp521r1",
            ]
        },
    )
    if not certificate_list_ret["result"]:
        hub.log.debug(
            f"Could not list AWS Certificates {certificate_list_ret['comment']}"
        )
        return {}
    certificate_list = certificate_list_ret["ret"].get("CertificateSummaryList")

    for certificate in certificate_list:
        resource_id = certificate.get("CertificateArn")
        ret = await hub.exec.boto3.client.acm.describe_certificate(
            ctx, CertificateArn=resource_id
        )
        if not ret["result"]:
            hub.log.debug(
                f"Could not describe AWS Certificate '{resource_id}' {ret['comment']}"
            )
            continue

        resource = ret["ret"].get("Certificate")

        convert_ret = (
            await hub.tool.aws.acm.conversion_utils.convert_raw_acm_to_present_async(
                ctx,
                raw_resource=resource,
                idem_resource_name=resource_id,
            )
        )
        if not convert_ret["result"]:
            hub.log.debug(
                f"Could not describe AWS Certificate '{resource_id}' {convert_ret['comment']}"
            )
            continue
        resource_translated = convert_ret.get("ret")
        result[resource_id] = {
            "aws.acm.certificate_manager.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource_translated.items()
            ]
        }

    return result



"""Exec module for managing Cloudwatch Event Rule."""
from typing import Dict

__func_alias__ = {"list_": "list"}


async def get(
    hub,
    ctx,
    name,
    resource_id: str,
    event_bus_name: str = None,
) -> Dict:
    """Get an event rule resource.

    Args:
        name(str):
            The name of the Idem state.

        resource_id(str):
            The name of the AWS CloudWatch Events rule to identify the resource.

        event_bus_name(str, Optional):
            The name or ARN of the event bus to associate with this rule.
            If you omit this, the default event bus is used.

    Returns:
        .. code-block:: python

            {"result": True|False, "comment": list, "ret": None|dict}

    Examples:
        Calling this exec module function from the cli with resource_id

        .. code-block:: bash

            idem exec aws.events.rule.get name="asg_1" resource_id="resource_id"

        Calling this exec module function from within a state module in pure python

        .. code-block:: python

            async def state_function(hub, ctx, name, resource_id):
                ret = await hub.exec.aws.events.rule.get(
                    ctx=ctx,
                    name=name,
                    resource_id=resource_id
                )

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: aws.events.rule.get
                - kwargs:
                    name: my_resources
                    resource_id: idem-test-events-rule48f593a7-e5ff-48ee-ab30-422241a7e693
    """
    result = dict(comment=[], ret=None, result=True)

    resource_ret = await hub.exec.boto3.client.events.describe_rule(
        ctx, Name=resource_id, EventBusName=event_bus_name
    )

    if not resource_ret["result"]:
        if "ResourceNotFoundException" in str(resource_ret["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.events.rule", name=name
                )
            )
            result["comment"] += list(resource_ret["comment"])
            return result
        result["comment"] += list(resource_ret["comment"])
        result["result"] = False
        return result

    if not resource_ret["ret"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type="aws.events.rule", name=name
            )
        )
        return result

    rule = resource_ret["ret"]
    tags = await hub.exec.boto3.client.events.list_tags_for_resource(
        ctx, ResourceARN=rule.get("Arn")
    )

    if not tags["result"]:
        result["result"] = False
        result["comment"] += list(tags["comment"])
        return result

    resource_converted = await hub.tool.aws.events.conversion_utils.convert_raw_cloud_watch_rule_to_present_async(
        ctx,
        raw_resource=rule,
        idem_resource_name=rule.get("Name"),
        tags=tags.get("ret").get("Tags"),
    )

    if not resource_converted["result"]:
        result["comment"] += list(resource_converted["comment"])
        result["result"] = False

    result["ret"] = resource_converted["ret"]

    return result


async def list_(hub, ctx, name: str = None) -> Dict:
    """List event rules from AWS.

    Args:
        name(str, Optional):
            The name of the Idem state.

    Returns:
        Dict[str, Any]

    Examples:
        Calling this exec module function from the cli

        .. code-block:: bash

            idem exec aws.events.rule.list name="my_resources"

        Calling this exec module function from within a state module in pure python.

        .. code-block:: python

            async def state_function(hub, ctx, name):
                await hub.exec.aws.events.rule.list(
                    ctx=ctx,
                    name=name
                )

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: aws.events.rule.list
                - kwargs:
                    name: my_resources
    """
    result = dict(comment=[], ret=None, result=True)
    ret = await hub.exec.boto3.client.events.list_rules(ctx)
    if not ret["result"]:
        result["comment"] = list(ret["comment"])
        result["result"] = False
        return result

    if not ret["ret"]["Rules"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.list_empty_comment(
                resource_type="aws.events.rule", name=name
            )
        )
        return result
    result["ret"] = []

    for rule in ret["ret"]["Rules"]:
        tags = await hub.exec.boto3.client.events.list_tags_for_resource(
            ctx, ResourceARN=rule.get("Arn")
        )

        if not tags["result"]:
            hub.log.warning(
                f"Could not retrieve tags for aws.events.rule '{rule.get('Name')}' "
                f"with error {tags['comment']}"
            )
        resource_converted = await hub.tool.aws.events.conversion_utils.convert_raw_cloud_watch_rule_to_present_async(
            ctx,
            raw_resource=rule,
            idem_resource_name=rule.get("Name"),
            tags=tags.get("ret").get("Tags") if tags.get("result") else None,
        )
        if not resource_converted["result"]:
            hub.log.warning(
                f"Could not convert aws.events.rule '{rule.get('Name')}' "
                f"with error {resource_converted['comment']}"
            )
        else:
            result["ret"].append(resource_converted["ret"])

    return result



"""
Exec module for managing Rds Db Proxys.
"""
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["soft_fail"]

__func_alias__ = {"list_": "list"}

create_waiter_acceptors = [
    {
        "matcher": "pathAll",
        "expected": "available",
        "state": "success",
        "argument": "DBProxies[].Status",
    },
    {
        "matcher": "pathAll",
        "expected": "creating",
        "state": "retry",
        "argument": "DBProxies[].Status",
    },
]

update_waiter_acceptors = [
    {
        "matcher": "pathAll",
        "expected": "available",
        "state": "success",
        "argument": "DBProxies[].Status",
    },
    {
        "matcher": "pathAll",
        "expected": "modifying",
        "state": "retry",
        "argument": "DBProxies[].Status",
    },
    {
        "matcher": "pathAll",
        "expected": "resetting-master-credentials",
        "state": "retry",
        "argument": "DBProxies[].Status",
    },
    {
        "matcher": "pathAll",
        "expected": "renaming",
        "state": "retry",
        "argument": "DBProxies[].Status",
    },
    {
        "matcher": "pathAll",
        "expected": "upgrading",
        "state": "retry",
        "argument": "DBProxies[].Status",
    },
]
delete_waiter_acceptors = [
    {
        "matcher": "error",
        "expected": "DBProxyNotFoundFault",
        "state": "success",
        "argument": "Error.Code",
    },
    {
        "matcher": "pathAll",
        "expected": "available",
        "state": "retry",
        "argument": "DBProxies[].Status",
    },
    {
        "matcher": "pathAll",
        "expected": "deleting",
        "state": "retry",
        "argument": "DBProxies[].Status",
    },
]


async def get(
    hub,
    ctx,
    resource_id: str,
    filters: List[
        make_dataclass("Filter", [("Name", str), ("Values", List[str])])
    ] = None,
    name: str = None,
) -> Dict[str, Any]:
    """

    Returns information about DB proxies.

    Args:
        resource_id(str): Db_proxy unique ID. The name of the DB proxy.

        filters(List[dict[str, Any]], Optional): This parameter is not currently supported. Defaults to None.

            * Name (str): The name of the filter. Filter names are case-sensitive.

            * Values (List[str]): One or more filter values. Filter values are case-sensitive.

        name(str, Optional): Idem name of the resource. Defaults to None.

    Returns:
        Dict[str, Any]

    Examples:
        Resource State:

        .. code-block:: sls

            unmanaged_resource:
              exec.run:
                - path: aws.rds.db_proxy.get
                - kwargs:
                  resource_id: value

        Exec call from the CLI:

        .. code-block:: bash

            idem exec aws.rds.db_proxy.get resource_id=value
    """

    result = dict(comment=[], ret=None, result=True)
    get_rds_db_proxy = await hub.exec.boto3.client.rds.describe_db_proxies(
        ctx=ctx, DBProxyName=resource_id, Filters=filters
    )
    # Case: Error
    if not get_rds_db_proxy["result"]:
        # Do not return success=false when it is not found.
        # Most of the resources would return "*NotFound*" type of exception when it is 404
        if "DBProxyNotFoundFault" in str(get_rds_db_proxy["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.rds.db_proxy", name=resource_id
                )
            )
            result["comment"].append(get_rds_db_proxy["comment"])
            return result

        result["comment"].append(get_rds_db_proxy["comment"])
        result["result"] = False
        return result

    # Case: Empty results
    if not get_rds_db_proxy["ret"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type="aws.rds.db_proxy", name=resource_id
            )
        )
        return result

    if len(get_rds_db_proxy["ret"]["DBProxies"]) > 1:
        result["comment"].append(
            hub.tool.aws.comment_utils.find_more_than_one(
                resource_id=resource_id, resource_type="aws.rds.db_proxy"
            )
        )

    raw_resource = get_rds_db_proxy["ret"]["DBProxies"][0]

    result[
        "ret"
    ] = await hub.tool.aws.rds.db_proxy.convert_raw_resource_to_present_async(
        ctx=ctx,
        resource_id=resource_id,
        raw_resource=raw_resource,
        idem_resource_name=name,
    )

    return result


async def list_(
    hub,
    ctx,
    resource_id: str = None,
    filters: List[
        make_dataclass("Filter", [("Name", str), ("Values", List[str])])
    ] = None,
    name: str = None,
) -> Dict[str, Any]:
    """
    Returns information about DB proxies.

    Args:
        resource_id(str, Optional): The name of the DB proxy. If you omit this parameter, the output includes information about all
            DB proxies owned by your Amazon Web Services account ID. Defaults to None.

        filters(List[dict[str, Any]], Optional): This parameter is not currently supported. Defaults to None.

            * Name (str): The name of the filter. Filter names are case-sensitive.

            * Values (List[str]): One or more filter values. Filter values are case-sensitive.

        name(str, Optional): Idem name of the resource. Defaults to None.

    Returns:
        Dict[str, Any]

    Examples:

        Resource State:

        .. code-block:: sls

            unmanaged_resources:
              exec.run:
                - path: aws.rds.db_proxy.list
                - kwargs:


        Exec call from the CLI:

        .. code-block:: bash

            idem exec aws.rds.db_proxy.list

        Describe call from the CLI:

        .. code-block:: bash

            $ idem describe aws.rds.db_proxy

    """

    result = dict(comment=[], ret=[], result=True)

    ret = await hub.exec.boto3.client.rds.describe_db_proxies(
        ctx=ctx, DBProxyName=resource_id, Filters=filters
    )

    if not ret["result"]:
        result["comment"].append(ret["comment"])
        result["result"] = False
        return result

    if not ret["ret"].get("DBProxies"):
        result["comment"].append(
            hub.tool.aws.comment_utils.list_empty_comment(
                resource_type="aws.rds.db_proxy", name=name
            )
        )
        return result

    for resource in ret["ret"]["DBProxies"]:
        result["ret"].append(
            await hub.tool.aws.rds.db_proxy.convert_raw_resource_to_present_async(
                ctx=ctx,
                resource_id=resource["DBProxyName"],
                raw_resource=resource,
                idem_resource_name=resource["DBProxyName"],
            )
        )
    return result


async def create(
    hub,
    ctx,
    db_proxy_name: str,
    engine_family: str,
    auth: List[
        make_dataclass(
            "UserAuthConfig",
            [
                ("Description", str, field(default=None)),
                ("UserName", str, field(default=None)),
                ("AuthScheme", str, field(default=None)),
                ("SecretArn", str, field(default=None)),
                ("IAMAuth", str, field(default=None)),
                ("ClientPasswordAuthType", str, field(default=None)),
            ],
        )
    ],
    role_arn: str,
    vpc_subnet_ids: List[str],
    vpc_security_group_ids: List[str] = None,
    require_tls: bool = None,
    idle_client_timeout: int = None,
    debug_logging: bool = None,
    tags: Dict[str, str] = None,
    name: str = None,
    timeout: make_dataclass(
        """Timeout configuration.""" "Timeout",
        [
            (
                "create",
                make_dataclass(
                    "CreateTimeout",
                    [
                        ("delay", int, field(default=None)),
                        ("max_attempts", int, field(default=None)),
                    ],
                ),
                field(default=None),
            ),
            (
                "update",
                make_dataclass(
                    "UpdateTimeout",
                    [
                        ("delay", int, field(default=None)),
                        ("max_attempts", int, field(default=None)),
                    ],
                ),
                field(default=None),
            ),
        ],
    ) = None,
) -> Dict[str, Any]:
    """
    Creates a new DB proxy.

    Args:
        db_proxy_name(str): The identifier for the proxy. This name must be unique for all proxies owned by your Amazon Web
            Services account in the specified Amazon Web Services Region. An identifier must begin with a
            letter and must contain only ASCII letters, digits, and hyphens; it can't end with a hyphen or
            contain two consecutive hyphens.

        engine_family(str): The kinds of databases that the proxy can connect to. This value determines which database
            network protocol the proxy recognizes when it interprets network traffic to and from the
            database. For Aurora MySQL, RDS for MariaDB, and RDS for MySQL databases, specify MYSQL. For
            Aurora PostgreSQL and RDS for PostgreSQL databases, specify POSTGRESQL. For RDS for Microsoft
            SQL Server, specify SQLSERVER.

        auth(List[dict[str, Any]]): The authorization mechanism that the proxy uses.

            * Description (str, Optional): A user-specified description about the authentication used by a proxy to log in as a specific
            database user.

            * UserName (str, Optional): The name of the database user to which the proxy connects.

            * AuthScheme (str, Optional): The type of authentication that the proxy uses for connections from the proxy to the underlying
            database.

            * SecretArn (str, Optional): The Amazon Resource Name (ARN) representing the secret that the proxy uses to authenticate to
            the RDS DB instance or Aurora DB cluster. These secrets are stored within Amazon Secrets
            Manager.

            * IAMAuth (str, Optional): Whether to require or disallow Amazon Web Services Identity and Access Management (IAM)
            authentication for connections to the proxy. The ENABLED value is valid only for proxies with
            RDS for Microsoft SQL Server.

            * ClientPasswordAuthType (str, Optional): The type of authentication the proxy uses for connections from clients.

        role_arn(str): The Amazon Resource Name (ARN) of the IAM role that the proxy uses to access secrets in Amazon
            Web Services Secrets Manager.

        vpc_subnet_ids(List[str]): One or more VPC subnet IDs to associate with the new proxy.

        vpc_security_group_ids(List[str], Optional): One or more VPC security group IDs to associate with the new proxy. Defaults to None.

        require_tls(bool, Optional): A Boolean parameter that specifies whether Transport Layer Security (TLS) encryption is required
            for connections to the proxy. By enabling this setting, you can enforce encrypted TLS
            connections to the proxy. Defaults to None.

        idle_client_timeout(int, Optional): The number of seconds that a connection to the proxy can be inactive before the proxy
            disconnects it. You can set this value higher or lower than the connection timeout limit for the
            associated database. Defaults to None.

        debug_logging(bool, Optional): Whether the proxy includes detailed information about SQL statements in its logs. This
            information helps you to debug issues involving SQL behavior or the performance and scalability
            of the proxy connections. The debug information includes the text of SQL statements that you
            submit through the proxy. Thus, only enable this setting when needed for debugging, and only
            when you have security measures in place to safeguard any sensitive information that appears in
            the logs. Defaults to None.

        tags(Dict[str, str], Optional): The tags to apply to the resource.

        name(str, Optional): Idem name of the resource. Defaults to None.

        timeout(dict, Optional):
            Timeout configuration for create/update of AWS DB Cluster.

            * create (*dict, Optional*):
                Timeout configuration for creating DB Cluster.

                * delay(*int, Optional*):
                    The amount of time in seconds to wait between attempts.

                * max_attempts(*int, Optional*):
                    Customized timeout configuration containing delay and max attempts.

            * update(*dict, Optional*):
                Timeout configuration for updating DB Cluster

                * delay(*int, Optional*):
                    The amount of time in seconds to wait between attempts.

                * max_attempts(*int, Optional*):
                    Customized timeout configuration containing delay and max attempts.

    Returns:
        Dict[str, Any]

    Examples:
        Using in a state:

        .. code-block:: sls

            resource_is_present:
              aws.rds.db_proxy.present:
                - db_proxy_name: value
                - engine_family: value
                - auth: value
                - role_arn: value
                - vpc_subnet_ids: value

        Exec call from the CLI:

        .. code-block:: bash

            idem exec aws.rds.db_proxy.create db_proxy_name=value, engine_family=value, auth=value, role_arn=value, vpc_subnet_ids=value
    """

    result = dict(comment=[], ret={}, result=True)

    desired_state = {
        k: v
        for k, v in locals().items()
        if k not in ("hub", "ctx", "kwargs", "result") and v is not None
    }

    resource_to_raw_input_mapping = {
        "db_proxy_name": "DBProxyName",
        "engine_family": "EngineFamily",
        "auth": "Auth",
        "role_arn": "RoleArn",
        "vpc_subnet_ids": "VpcSubnetIds",
        "vpc_security_group_ids": "VpcSecurityGroupIds",
        "require_tls": "RequireTLS",
        "idle_client_timeout": "IdleClientTimeout",
        "debug_logging": "DebugLogging",
    }

    payload = {}
    for key, value in desired_state.items():
        if key in resource_to_raw_input_mapping.keys() and value is not None:
            payload[resource_to_raw_input_mapping[key]] = desired_state.get(key)

    if tags:
        payload["Tags"] = hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags)

    ret = await hub.exec.boto3.client.rds.create_db_proxy(ctx, **payload)

    result["result"] = ret["result"]
    if not result["result"]:
        result["comment"] += ret["comment"]
        return result
    raw_resource = ret["ret"].get("DBProxy", {})
    result["ret"]["resource_id"] = raw_resource.get("DBProxyName")
    result["ret"]["name"] = name

    waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
        default_delay=20,
        default_max_attempts=40,
        timeout_config=timeout.get("create") if timeout else None,
    )
    cluster_waiter = hub.tool.boto3.custom_waiter.waiter_wrapper(
        name="DBProxyCreated",
        operation="DescribeDBProxies",
        argument=["DBProxies[].Status"],
        acceptors=create_waiter_acceptors,
        client=await hub.tool.boto3.client.get_client(ctx, "rds"),
    )
    try:
        await hub.tool.boto3.client.wait(
            ctx,
            "rds",
            "DBProxyCreated",
            cluster_waiter,
            DBProxyName=result["ret"]["resource_id"],
            WaiterConfig=waiter_config,
        )
    except Exception as e:
        result["comment"] += (str(e),)
        result["result"] = False
        return result

    result["comment"] += hub.tool.aws.comment_utils.create_comment(
        resource_type="aws.rds.db_proxy", name=name
    )

    return result


async def update(
    hub,
    ctx,
    resource_id: str,
    new_db_proxy_name: str = None,
    auth: List[
        make_dataclass(
            "UserAuthConfig",
            [
                ("Description", str, field(default=None)),
                ("UserName", str, field(default=None)),
                ("AuthScheme", str, field(default=None)),
                ("SecretArn", str, field(default=None)),
                ("IAMAuth", str, field(default=None)),
                ("ClientPasswordAuthType", str, field(default=None)),
            ],
        )
    ] = None,
    require_tls: bool = None,
    idle_client_timeout: int = None,
    debug_logging: bool = None,
    role_arn: str = None,
    security_groups: List[str] = None,
    name: str = None,
    tags: Dict[str, str] = None,
    timeout: make_dataclass(
        """Timeout configuration.""" "Timeout",
        [
            (
                "create",
                make_dataclass(
                    "CreateTimeout",
                    [
                        ("delay", int, field(default=None)),
                        ("max_attempts", int, field(default=None)),
                    ],
                ),
                field(default=None),
            ),
            (
                "update",
                make_dataclass(
                    "UpdateTimeout",
                    [
                        ("delay", int, field(default=None)),
                        ("max_attempts", int, field(default=None)),
                    ],
                ),
                field(default=None),
            ),
        ],
    ) = None,
) -> Dict[str, Any]:
    """
    Changes the settings for an existing DB proxy.

    Args:

        resource_id(str): Db_proxy unique ID.

        new_db_proxy_name(str, Optional): The new identifier for the DBProxy. An identifier must begin with a letter and must contain only
            ASCII letters, digits, and hyphens; it can't end with a hyphen or contain two consecutive
            hyphens. Defaults to None.

        auth(List[dict[str, Any]], Optional): The new authentication settings for the DBProxy. Defaults to None.

            * Description (str, Optional): A user-specified description about the authentication used by a proxy to log in as a specific
            database user.

            * UserName (str, Optional): The name of the database user to which the proxy connects.

            * AuthScheme (str, Optional): The type of authentication that the proxy uses for connections from the proxy to the underlying
            database.

            * SecretArn (str, Optional): The Amazon Resource Name (ARN) representing the secret that the proxy uses to authenticate to
            the RDS DB instance or Aurora DB cluster. These secrets are stored within Amazon Secrets
            Manager.

            * IAMAuth (str, Optional): Whether to require or disallow Amazon Web Services Identity and Access Management (IAM)
            authentication for connections to the proxy. The ENABLED value is valid only for proxies with
            RDS for Microsoft SQL Server.

            * ClientPasswordAuthType (str, Optional): The type of authentication the proxy uses for connections from clients.

        require_tls(bool, Optional): Whether Transport Layer Security (TLS) encryption is required for connections to the proxy. By
            enabling this setting, you can enforce encrypted TLS connections to the proxy, even if the
            associated database doesn't use TLS. Defaults to None.

        idle_client_timeout(int, Optional): The number of seconds that a connection to the proxy can be inactive before the proxy
            disconnects it. You can set this value higher or lower than the connection timeout limit for the
            associated database. Defaults to None.

        debug_logging(bool, Optional): Whether the proxy includes detailed information about SQL statements in its logs. This
            information helps you to debug issues involving SQL behavior or the performance and scalability
            of the proxy connections. The debug information includes the text of SQL statements that you
            submit through the proxy. Thus, only enable this setting when needed for debugging, and only
            when you have security measures in place to safeguard any sensitive information that appears in
            the logs. Defaults to None.

        role_arn(str, Optional): The Amazon Resource Name (ARN) of the IAM role that the proxy uses to access secrets in Amazon
            Web Services Secrets Manager. Defaults to None.

        security_groups(List[str], Optional): The new list of security groups for the DBProxy. Defaults to None.

        name(str, Optional): Idem name of the resource. Defaults to None.

        tags(Dict[str, str], Optional): The tags to apply to the resource.

        timeout(dict, Optional):
            Timeout configuration for create/update of AWS DB Cluster.

            * create (*dict, Optional*):
                Timeout configuration for creating DB Cluster.

                * delay(*int, Optional*):
                    The amount of time in seconds to wait between attempts.

                * max_attempts(*int, Optional*):
                    Customized timeout configuration containing delay and max attempts.

            * update(*dict, Optional*):
                Timeout configuration for updating DB Cluster

                * delay(*int, Optional*):
                    The amount of time in seconds to wait between attempts.

                * max_attempts(*int, Optional*):
                    Customized timeout configuration containing delay and max attempts.

    Returns:
        Dict[str, Any]

    Examples:
        Using in a state:

        .. code-block:: sls

            resource_is_present:
              aws.rds.db_proxy.present:
                - db_proxy_name: value
                - tags: value
                - resource_id: value

        Exec call from the CLI:

        .. code-block:: bash

            idem exec aws.rds.db_proxy.update db_proxy_name=value, tags=value, resource_id=value
    """

    result = dict(comment=[], ret={}, result=True)

    desired_state = {
        k: v
        for k, v in locals().items()
        if k not in ("hub", "ctx", "kwargs", "result") and v is not None
    }

    payload = {}

    resource_to_raw_input_mapping = {
        "resource_id": "DBProxyName",
        "new_db_proxy_name": "NewDBProxyName",
        "auth": "Auth",
        "require_tls": "RequireTLS",
        "idle_client_timeout": "IdleClientTimeout",
        "debug_logging": "DebugLogging",
        "role_arn": "RoleArn",
        "security_groups": "SecurityGroups",
    }

    for key, value in desired_state.items():
        if key in resource_to_raw_input_mapping.keys() and value is not None:
            payload[resource_to_raw_input_mapping[key]] = desired_state.get(key)

    if payload:
        ret = await hub.exec.boto3.client.rds.modify_db_proxy(ctx, **payload)
        if not ret["result"]:
            result["result"] = False
            result["comment"].append(
                f"Could not update aws.rds.db_proxy '{name}'",
            )
            result["comment"] += ret["comment"]
            return result

        raw_resource = ret["ret"].get("DBProxy", {})
        result["comment"] += hub.tool.aws.comment_utils.update_comment(
            resource_type="aws.rds.db_proxy", name=name
        )

        get_tags_ret = await hub.tool.aws.rds.tag.get_tags_for_resource(
            ctx, resource_arn=raw_resource.get("DBProxyArn")
        )

        if get_tags_ret["result"]:
            current_tags = get_tags_ret.get("ret", {})
            update_tags_ret = await hub.tool.aws.rds.tag.update_tags(
                ctx,
                resource_arn=raw_resource.get("DBProxyArn"),
                old_tags=current_tags,
                new_tags=tags,
            )
            if not update_tags_ret["result"]:
                result["result"] = False
                result["comment"] += update_tags_ret["comment"]
                return result

        result["ret"]["resource_id"] = raw_resource.get("DBProxyName")
        result["ret"]["name"] = name

        waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
            default_delay=20,
            default_max_attempts=40,
            timeout_config=timeout.get("update") if timeout else None,
        )
        cluster_waiter = hub.tool.boto3.custom_waiter.waiter_wrapper(
            name="DBProxyUpdated",
            operation="DescribeDBProxies",
            argument=["DBProxies[].Status"],
            acceptors=update_waiter_acceptors,
            client=await hub.tool.boto3.client.get_client(ctx, "rds"),
        )
        try:
            await hub.tool.boto3.client.wait(
                ctx,
                "rds",
                "DBProxyUpdated",
                cluster_waiter,
                DBProxyName=result["ret"]["resource_id"],
                WaiterConfig=waiter_config,
            )
        except Exception as e:
            result["comment"] += (str(e),)
            result["result"] = False
            return result

    return result


async def delete(
    hub,
    ctx,
    resource_id: str,
    name: str = None,
    timeout: make_dataclass(
        """Timeout configuration.""" "Timeout",
        [
            (
                "delete",
                make_dataclass(
                    "DeleteTimeout",
                    [
                        ("delay", int, field(default=None)),
                        ("max_attempts", int, field(default=None)),
                    ],
                ),
                field(default=None),
            ),
        ],
    ) = None,
) -> Dict[str, Any]:
    """
    Deletes an existing DB proxy.

    Args:
        resource_id(str): Db_proxy unique ID.

        name(str, Optional): Idem name of the resource. Defaults to None.

        timeout(dict, Optional):
            Timeout configuration for delete of AWS RDS DB Proxy.

            * delete (*dict, Optional*):
                Timeout configuration for deleting DB Cluster.

                * delay(*int, Optional*):
                    The amount of time in seconds to wait between attempts.

                * max_attempts(*int, Optional*):
                    Customized timeout configuration containing delay and max attempts.


    Returns:
        Dict[str, Any]

    Examples:
        Resource State:

        .. code-block:: sls

            resource_is_absent:
              aws.rds.db_proxy.absent:
                - db_proxy_name: value
                - resource_id: value

        Exec call from the CLI:

        .. code-block:: bash

            idem exec aws.rds.db_proxy.delete db_proxy_name=value, resource_id=value
    """

    result = dict(comment=[], ret=None, result=True)

    delete_ret = await hub.exec.boto3.client.rds.delete_db_proxy(
        ctx, **{"DBProxyName": resource_id}
    )

    result["result"] = delete_ret["result"]

    if not result["result"]:
        result["comment"] = delete_ret["comment"]
        result["result"] = False
        return result

    result["comment"] += hub.tool.aws.comment_utils.delete_comment(
        resource_type="aws.rds.db_proxy", name=name
    )

    waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
        default_delay=20,
        default_max_attempts=40,
        timeout_config=timeout.get("delete") if timeout else None,
    )
    cluster_waiter = hub.tool.boto3.custom_waiter.waiter_wrapper(
        name="DBProxyDeleted",
        operation="DescribeDBProxies",
        argument=["DBProxies[].Status"],
        acceptors=delete_waiter_acceptors,
        client=await hub.tool.boto3.client.get_client(ctx, "rds"),
    )
    try:
        await hub.tool.boto3.client.wait(
            ctx,
            "rds",
            "DBProxyDeleted",
            cluster_waiter,
            DBProxyName=resource_id,
            WaiterConfig=waiter_config,
        )
    except Exception as e:
        result["comment"] += (str(e),)
        result["result"] = False
        return result

    return result



import datetime
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

from dict_tools.typing import Computed

__contracts__ = ["resource", "allow_sync_sls_name_and_name_tag"]
__reconcile_wait__ = {"static": {"wait_in_seconds": 5}}

TREQ = {
    "present": {
        "require": [
            "aws.ec2.network_interface.present",
            "aws.ec2.vpc.present",
            "aws.ec2.subnet.present",
            "aws.iam.role.present",
        ],
    },
}


async def present(
    hub,
    ctx,
    name: str,
    image_id: str,
    *,
    resource_id: str = None,
    # From DescribeInstances
    instance_type: str = None,
    ebs_optimized: bool = None,
    kernel_id: str = None,
    subnet_id: str = None,
    security_group_ids: List[str] = None,
    network_interfaces: List[
        make_dataclass(
            "NetworkInterfaceAttachment",
            [
                ("device_index", int),
                ("network_interface_id", str, field(default=None)),
                ("network_card_index", int, field(default=None)),
                ("subnet_id", Computed[str], field(default=None)),
                ("public_ip_address", Computed[str], field(default=None)),
                ("private_ip_address", Computed[str], field(default=None)),
                ("mac_address", Computed[str], field(default=None)),
            ],
        )
    ] = None,
    monitoring_enabled: bool = None,
    source_dest_check: bool = None,
    running: bool = True,
    private_ip_address: str = None,
    owner_id: str = None,
    # Placement Options
    availability_zone: str = None,
    affinity: str = None,
    group_name: str = None,
    partition_number: int = None,
    host_id: str = None,
    tenancy: str = None,
    spread_domain: str = None,
    host_resource_group_arn: str = None,
    # From launchTemplate
    user_data: str = None,
    disable_api_termination: bool = None,
    ram_disk_id: str = None,
    tags: Dict[str, str] = None,
    iam_instance_profile_arn: str = None,
    instance_initiated_shutdown_behavior: str = None,
    elastic_inference_accelerators: List[
        make_dataclass(
            "ElasticInferenceAccelerator",
            [("Type", str), ("Count", int, field(default=None))],
        )
    ] = None,
    elastic_gpu_specifications: List[
        make_dataclass("ElasticGpuSpecification", [("Type", str)])
    ] = None,
    auto_recovery_enabled: bool = None,
    sriov_net_support: str = None,
    key_name: str = None,
    # Enclave options
    nitro_enclave_enabled: bool = None,
    # Can only be changed on initial creation of an instance as far as we know so far
    client_token: str = None,
    product_codes: List[Dict[str, str]] = None,
    reservation_id: str = None,
    block_device_mappings: List[
        make_dataclass(
            "BlockDeviceMapping",
            [
                ("device_name", str),
                ("volume_id", str, field(default=None)),
                ("delete_on_termination", bool, field(default=False)),
                # computed volume properties
                ("size", Computed[int], field(default=None)),
                ("type", Computed[str], field(default=None)),
                ("encrypted", Computed[bool], field(default=None)),
            ],
        )
    ] = None,
    license_arns: List[str] = None,
    hibernation_enabled: bool = None,
    # Instance Market Options
    market_type: str = None,
    # Spot Options
    max_price: str = None,
    spot_instance_type: str = None,
    block_duration_minutes: int = None,
    valid_until: str = None,
    instance_interruption_behavior: str = None,
    # Credit Specification
    cpu_credits: str = None,
    cpu_core_count: int = None,
    cpu_threads_per_core: int = None,
    # Metadata options
    http_tokens: str = None,
    http_put_response_hop_limit: int = None,
    http_endpoint_enabled: bool = None,
    http_protocol_ipv6_enabled: bool = None,
    metadata_tags_enabled: bool = None,
    # Private DNS Name options
    hostname_type: str = None,
    enable_resource_name_dns_a_record: bool = None,
    enable_resource_name_dns_aaaa_record: bool = None,
    # Capacity Reservation Options
    capacity_reservation_preference: str = None,
    instance_requirements: Dict[str, Any] = None,
    # computed properties
    instance_state: Computed[str] = None,
    public_ip_address: Computed[str] = None,
    root_device_name: Computed[str] = None,
    # idem-heist options
    bootstrap: List[Dict[str, str]] = None,
    **kwargs,
) -> Dict[str, Any]:
    """Launches an instance using an AMI for which you have permissions.

    You can specify a number of options, or leave the default options.

    The following rules apply:

        - [EC2-VPC] If you don't specify a subnet ID, we choose a default subnet from your default VPC for you.
            If you don't have a default VPC, you must specify a subnet ID in the request.
        - [EC2-Classic] If don't specify an Availability Zone, we choose one for you.
            Some instance types must be launched into a VPC.
            If you do not have a default VPC, or if you do not specify a subnet ID, the request fails.
            For more information, see Instance types available only in a VPC.
        - [EC2-VPC] All instances have a network interface with a primary private IPv4 address.
            If you don't specify this address, we choose one from the IPv4 range of your subnet.
            Not all instance types support IPv6 addresses.
            For more information, see Instance types.
        - If you don't specify a security group ID, we use the default security group.
            For more information, see Security groups.
        - If any of the AMIs have a product code attached for which the user has not subscribed, the request fails.
        - You can create a launch template, which is a resource that contains the parameters to launch an instance.
            You can specify the launch template instead of specifying the launch parameters.
            An instance is ready for you to use when it's in the running state.
        - You can tag instances and EBS volumes during launch, after launch, or both.
        - Linux instances have access to the public key of the key pair at boot.
            You can use this key to provide secure access to the instance.
            Amazon EC2 public images use this feature to provide secure access without passwords.
            For more information, see Key pairs.

    Args:
        name(str):
            An Idem name of the resource.

        resource_id(str):
            AWS Ec2 Instance ID.

        image_id(str):
            The ID of an AMI.

        instance_type(str, Optional):
            The instance type to use for this instance on creation.

        ebs_optimized(bool, Optional):
            Indicates whether the instance is optimized ofr Amazon EBS I/O.

        kernel_id(str, Optional):
            The kernel associated with this instance, if applicable.

        subnet_id(str, Optional):
            The ID of the subnet in which the instance is running.

        security_group_ids(List[str], Optional):
            The IDs of the security groups. If you specify a network interface, you must specify any security
            groups as part of the network interface. Default: Amazon EC2 uses the default security group.

        network_interfaces(list[dict[str, Any]], Optional):
            The network interfaces to associate with the instance. This manages the network interface attachment
            to the instance, network interfaces themselves are managed in a separate state. Defaults to None.

            * network_interface_id (str, Optional):
                The ID of the network interface. If you are creating a Spot Fleet, omit this parameter because
                you can’t specify a network interface ID in a launch specification.

            * device_index (int, Optional):
                The position of the network interface in the attachment order. A primary network interface has a
                device index of 0. If you specify a network interface when launching an instance, you must
                specify the device index.

            * network_card_index (int, Optional):
                The index of the network card. Some instance types support multiple network cards. The primary
                network interface must be assigned to network card index 0. The default is network card index 0.
                If you are using RequestSpotInstances to create Spot Instances, omit this parameter because you
                can’t specify the network card index when using this API. To specify the network card index, use
                RunInstances.

            * subnet_id (Computed[str]):
                The ID of the subnet.

            * public_ip_address (Computed[str]):
                The public IP address or Elastic IP address bound to the network interface.

            * private_ip_address (Computed[str]):
                The IPv4 address of the network interface within the subnet.

            * mac_address (Computed[str]):
                The MAC address.

        monitoring_enabled(bool, Optional):
            Indicates whether detailed monitoring is enabled.

        source_dest_check(bool, Optional):
            Indicates whether source/destination checking is enabled.

        running(bool, Optional):
            Indicates whether the instance should be in the "running" state.

        private_ip_address(str, Optional):
            The Ipv4 address of the network interface within the subnet.

        owner_id(str, Optional):
            The ID of the AWS account that owns the reservation.

        availability_zone(str, Optional):
            The Availability Zone of the instance.

        affinity(str, Optional):
            The affinity setting for the instance on the Dedicated Host.

        group_name(str, Optional):
            The affinity setting for the instance on the Dedicated Host.

        partition_number(int, Optional):
            The number of the partition that the instance is in.

        host_id(str, Optional):
            The ID of the Dedicated Host on which the instance resides.

        tenancy(str, Optional):
            The tenancy of the instance (if the instance is running in a VPC). An instance with a tenancy of dedicated runs on single-tenant hardware.

        spread_domain(str, Optional):
            Not yet documented by AWS.

        host_resource_group_arn(str, Optional):
            The ARN of the host resource group in which to launch the instances.

        user_data(str, Optional):
            The user data for the instance.

        disable_api_termination(bool, Optional):
            Indicates that an instance cannot be terminated using the Amazon Ec2 console, command line tool, or API.

        ram_disk_id(str, Optional):
            The ID of the RAM disk, if applicable.

        tags(dict, Optional, Optional):
            The tags to apply to the resource. Defaults to None.

        iam_instance_profile_arn(str, Optional):
            The IAM instance profile ARN.

        instance_initiated_shutdown_behavior(str, Optional): Indicates whether an instance stops or terminates when you initiate
            shutdown from the instance (using the operating system command for system shutdown).

        elastic_inference_accelerators(list[dict[str, Any]], Optional):
            An elastic inference accelerator to associate with the instance. Elastic inference accelerators
            are a resource you can attach to your Amazon EC2 instances to accelerate your Deep Learning (DL)
            inference workloads. You cannot specify accelerators from different generations in the same
            request. Defaults to None.

            * Type (str):
                The type of elastic inference accelerator. The possible values are eia1.medium, eia1.large,
                eia1.xlarge, eia2.medium, eia2.large, and eia2.xlarge.

            * Count (int, Optional):
                The number of elastic inference accelerators to attach to the instance.  Default: 1

        elastic_gpu_specifications(list[dict[str, Any]], Optional):
            An elastic GPU to associate with the instance. An Elastic GPU is a GPU resource that you can
            attach to your Windows instance to accelerate the graphics performance of your applications. For
            more information, see Amazon EC2 Elastic GPUs in the Amazon EC2 User Guide. Defaults to None.

            * Type (str): The type of Elastic Graphics accelerator. For more information about the values to specify for
                Type, see Elastic Graphics Basics, specifically the Elastic Graphics accelerator column, in the
                Amazon Elastic Compute Cloud User Guide for Windows Instances.

        auto_recovery_enabled(bool, Optional):
            Disables the automatic recovery behavior of your instance or sets it to default.

        sriov_net_support(str, Optional):
            Specifies whether enhanced networking with the Intel 82599 Virtual Function interface is enabled.

        key_name(str, Optional):
            The name of the keypair.

        nitro_enclave_enabled(bool, Optional):
            Indicates whether the instance is enabled for AWS Nitro Enclaves.

        client_token(str, Optional):
            The idempotency token for the instance.

        root_device_name(str, Optional):
            The device name of the root device (for example, /dev/sda1).

        product_codes(list[dict[str, str]]:
            The product codes attached to the instance, if applicable.

        reservation_id(str, Optional):
            The ID of the reservation

        block_device_mappings(list[dict[str, Any]], Optional):
            The block device mapping, which defines the EBS volumes and instance store volumes to attach to
            the instance at launch. For more information, see Block device mappings in the Amazon EC2 User
            Guide. Defaults to None.

            * device_name (str, Optional):
                The device name (for example, /dev/sdh or xvdh).

            * volume_id: (str):
                The resource_id of the volume to attach to the instance.

            * virtual_name (str, Optional):
                The virtual device name (ephemeralN). Instance store volumes are numbered starting from 0. An
                instance type with 2 available instance store volumes can specify mappings for ephemeral0 and
                ephemeral1. The number of available instance store volumes depends on the instance type. After
                you connect to the instance, you must mount the volume. NVMe instance store volumes are
                automatically enumerated and assigned a device name. Including them in your block device mapping
                has no effect. Constraints: For M3 instances, you must specify instance store volumes in the
                block device mapping for the instance. When you launch an M3 instance, we ignore any instance
                store volumes specified in the block device mapping for the AMI.

            * delete_on_termination (bool, Optional):
                Indicates whether the EBS volume is deleted on instance termination. For more information, see
                Preserving Amazon EBS volumes on instance termination in the Amazon EC2 User Guide.

        license_arns(list[str], Optional):
            The license configuration arns.

        hibernation_enabled(bool, Optional):
            Indicates whether the instance is configured for hibernation.

        market_type(str, Optional):
            The market (purchasing) option for the instance.

        max_price(str, Optional):
            The maximum hourly price you're willing to pay for the Spot Instances.

        spot_instance_type(str, Optional):
            The Spot Instance request type. Persistent Spot Instance requests are only supported when the instance
            interruption behavior is either hibernate or stop.

        block_duration_minutes(int, Optional):
            Deprecated.

        valid_until(str, Optional):
            The end date of the request, in UTC format (YYYY -MM -DD T*HH* :MM :SS Z). Supported only for persistent requests.

        instance_interruption_behavior(str, Optional):
            The behavior when a Spot Instance is interrupted. The default is terminate.

        cpu_credits(str, Optional):
            The credit option for CPU usage of a T2, T3, or T3a instance. Valid values are standard and unlimited.

        cpu_core_count(int, Optional):
            The number of CPU cores for the instance.

        cpu_threads_per_core(int, Optional):
            The number of threads per CPU core. To disable multithreading for the instance, specify a value of1 .
            Otherwise, specify the default value of 2.

        http_tokens(str, Optional):
            The state of token usage for your instance metadata requests. If the state is optional, you can
            choose to retrieve instance metadata with or without a signed token header on your request. If
            you retrieve the IAM role credentials without a token, the version 1.0 role credentials are
            returned. If you retrieve the IAM role credentials using a valid signed token, the version 2.0
            role credentials are returned. If the state is required, you must send a signed token header
            with any instance metadata retrieval requests. In this state, retrieving the IAM role
            credentials always returns the version 2.0 credentials; the version 1.0 credentials are not
            available. Default: Optional.

        http_put_response_hop_limit(int, Optional):
            The desired HTTP PUT response hop limit for instance metadata requests. The larger the number,
            the further instance metadata requests can travel. Default: 1 Possible values: Integers from 1
            to 64.

        http_endpoint_enabled(bool, Optional):
            Enables or disables the HTTP metadata endpoint on your instances. If you specify a value of
            disabled, you cannot access your instance metadata. Default: enabled.

        http_protocol_ipv6_enabled(str, Optional):
            Enables or disables the IPv6 endpoint for the instance metadata service.

        metadata_tags_enabled(bool, Optional):
            Set to enabled to allow access to instance tags from the instance metadata. Set to disabled to
            turn off access to instance tags from the instance metadata. For more information, see Work with
            instance tags using the instance metadata. Default: disabled.

        hostname_type(str, Optional):
            The type of hostname for EC2 instances.
            For IPv4 only subnets, an instance DNS name must be based on the instance IPv4 address.
            For IPv6 only subnets, an instance DNS name must be based on the instance ID.
            For dual-stack subnets, you can specify whether DNS names use the instance IPv4 address or the instance ID.

        enable_resource_name_dns_a_record(bool, Optional):
            Indicates whether to respond to DNS queries for instance hostnames with DNS A records.

        enable_resource_name_dns_aaaa_record(bool, Optional):
            Indicates whether to respond to DNS queries for instance hostnames with DNS A records.

        capacity_reservation_preference(str, Optional):
            Indicates the instance's Capacity Reservation preferences.

        instance_requirements(dict[str, Any], Optional):
            The attributes for the instance type. When you specify instance attributes, Amazon EC2 will identify
            instance types with these attributes.

        instance_state(Computed[str]):
            The reported state of the instance, possible values are:
            pending, running, shutting-down, terminated, stopping, stopped

        public_ip_address(Computed[str]):
            The public IPv4 address, or the Carrier IP address assigned to the instance, if applicable.
            A Carrier IP address only applies to an instance launched in a subnet associated with a Wavelength Zone.

        bootstrap(list[dict[str, Any]], Optional):
            Bootstr options for provisioning an instance with "heist".

    Request Syntax:
       .. code-block:: sls

          [instance-resource-id]:
            aws.ec2.instance.present:
              - name: "string"
              - image_id: "string"
              - tags:
                  - Key: "string"
                    Value: "string"
              - bootstrap:
                    - heist_manager: "string"
                      artifact_version: "string"

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            default_vpc:
              exec.run:
                - path: aws.ec2.vpc.get
                - kwargs:
                    default: "true"

            default_subnet:
              exec.run:
                - path: aws.ec2.subnet.get
                - kwargs:
                    filters:
                      - name: vpc-id
                        values:
                          - ${exec:default_vpc:resource_id}

            small_image:
              exec.run:
                - path: aws.ec2.ami.get
                - kwargs:
                    owners:
                      - amazon
                    most_recent: True
                    filters:
                      - name: image-type
                        values:
                          - machine
                      - name: state
                        values:
                          - available
                      - name: hypervisor
                        values:
                          - xen
                      - name: architecture
                        values:
                          - x86_64
                      - name: root-device-type
                        values:
                          - ebs
                      - name: virtualization-type
                        values:
                          - hvm

            small_instance_type:
              exec.run:
                - path: aws.ec2.instance_type.get
                - kwargs:
                    filters:
                      - name: instance-type
                        values:
                          - '*.nano'
                      - name: hypervisor
                        values:
                          - xen
                      - name: processor-info.supported-architecture
                        values:
                          - x86_64

            my_network_interface:
              aws.ec2.network_interface.present:
                - subnet_id: ${exec:default_subnet:resource_id}

            create_instance:
              aws.ec2.instance.present:
                - image_id: ${exec:small_image:resource_id}
                - instance_type: ${exec:small_instance_type:resource_id}
                - client_token: optional-idempotence-token
                - network_interfaces:
                  - network_interface_id: ${aws.ec2.network_interface:my_network_interface:resource_id}
                    device_index: 0
                    network_card_index: 0
                - tags:
                    Name: test-idem-cloud-demo
                - bootstrap:
                  - heist_manager: salt.minion
    """
    # Convert a list of dictionaries to a plain dictionary before calculating desired_state
    if isinstance(tags, list):
        tags = hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags)

    # Due to a bug, the default value populated to ESM is {}, for idem-aws versions <= 3.0.0
    if iam_instance_profile_arn == {}:
        iam_instance_profile_arn = None

    # Get all the parameters passed to this function as a single dictionary
    desired_state = {
        k: v
        for k, v in locals().items()
        if k not in ("hub", "ctx", "kwargs") and v is not None
    }

    result = dict(comment=[], old_state=None, new_state=None, name=name, result=True)

    current_state = None
    if resource_id:
        get = await hub.exec.aws.ec2.instance.get(
            ctx, name=name, resource_id=resource_id
        )
        current_state = result["old_state"] = get.ret

        if not result["old_state"]:
            result["comment"] += [
                f"Could not find instance for '{name}' with existing id '{resource_id}'"
            ]
            return result

        result["comment"] += [f"Instance '{name}' already exists"]
    elif client_token:
        get = await hub.exec.aws.ec2.instance.get(
            ctx, name=name, filters=[{"Name": "client-token", "Values": [client_token]}]
        )

        if get.result is True:
            if get.ret is None:
                result["comment"] += [
                    f"Could not find Instance for '{name}' with existing idempotence token '{client_token}'"
                ]
            else:
                current_state = result["old_state"] = get.ret
                resource_id = current_state.resource_id
                result["comment"] += [f"Instance '{name}' already exists"]

    if not resource_id:
        if ctx.test:
            result["new_state"] = hub.tool.aws.ec2.instance.state.test(**desired_state)
            result["comment"] += [f"Would create aws.ec2.instance '{name}'"]
            return result

        # Prepare options that are grouped in the request

        # Prepare instance market options
        spot_options = {}
        if max_price:
            spot_options["MaxPrice"] = max_price
        if spot_instance_type:
            spot_options["SpotInstanceType"] = spot_instance_type
        if block_duration_minutes:
            spot_options["BlockDurationMinutes"] = block_duration_minutes
        if valid_until:
            spot_options["ValidUntil"] = datetime.datetime(valid_until)
        if instance_initiated_shutdown_behavior:
            spot_options[
                "InstanceInterruptionBehavior"
            ] = instance_interruption_behavior
        if market_type:
            instance_market_options = {
                "MarketType": market_type,
                "SpotOptions": spot_options,
            }
        elif spot_options:
            instance_market_options = {"SpotOptions": spot_options}
        else:
            instance_market_options = None

        # Prepare Metadata options
        metadata_options = {}
        if http_tokens:
            metadata_options["HttpTokens"] = http_tokens
        if http_put_response_hop_limit:
            metadata_options["HttpPutResponseHopLimit"] = http_put_response_hop_limit
        if http_endpoint_enabled is not None:
            metadata_options["HttpEndpoint"] = (
                "enabled" if http_endpoint_enabled else "disabled"
            )
        if http_protocol_ipv6_enabled is not None:
            metadata_options["HttpProtocolIpv6"] = (
                "enabled" if http_protocol_ipv6_enabled else "disabled"
            )
        if metadata_tags_enabled is not None:
            metadata_options["InstanceMetadataTags"] = (
                "enabled" if metadata_tags_enabled else "disabled"
            )

        # Prepare Private DNS Name Options
        private_dns_name_options = {}
        if hostname_type:
            private_dns_name_options["HostnameType"] = hostname_type
        if enable_resource_name_dns_a_record is not None:
            private_dns_name_options[
                "EnableResourceNameDnsARecord"
            ] = enable_resource_name_dns_a_record
        if enable_resource_name_dns_aaaa_record is not None:
            private_dns_name_options[
                "EnableResourceNameDnsAAAARecord"
            ] = enable_resource_name_dns_aaaa_record

        # Prepare Cpu Options
        cpu_options = {}
        if cpu_core_count:
            cpu_options["CoreCount"] = cpu_core_count
        if cpu_threads_per_core:
            cpu_options["ThreadsPerCore"] = cpu_threads_per_core

        # Prepare Placement Options
        placement = {}
        if availability_zone:
            placement["AvailabilityZone"] = availability_zone
        if affinity:
            placement["Affinity"] = affinity
        if group_name:
            placement["GroupName"] = group_name
        if partition_number:
            placement["PartitionNumber"] = partition_number
        if host_id:
            placement["HostId"] = host_id
        if tenancy:
            placement["Tenancy"] = tenancy
        if spread_domain:
            placement["SpreadDomain"] = spread_domain
        if host_resource_group_arn:
            placement["HostResourceGroupArn"] = host_resource_group_arn

        # Prepare TagSpecifications
        if tags:
            tag_specifications = [
                {
                    "ResourceType": "instance",
                    "Tags": [{"Key": k, "Value": v} for k, v in tags.items()],
                }
            ]
        else:
            tag_specifications = None

        # Prepare Maintenance Options
        maintenance_options = {}
        if auto_recovery_enabled is not None:
            maintenance_options["AutoRecovery"] = (
                "disabled" if auto_recovery_enabled is False else "default"
            )

        # Prepare Hibernation Options
        hibernation_options = {}
        if hibernation_enabled is not None:
            hibernation_options["Configured"] = hibernation_enabled

        # Prepare Credit Specifications
        credit_specifications = {}
        if cpu_credits is not None:
            credit_specifications["CpuCredits"] = cpu_credits

        # Prepare Enclave Options
        enclave_options = {}
        if nitro_enclave_enabled:
            enclave_options["Enabled"] = nitro_enclave_enabled

        # Prepare Monitoring Options
        monitoring_options = {}
        if monitoring_enabled:
            monitoring_options["Enabled"] = monitoring_enabled

        # Prepare Capacity Reservation Specification
        capacity_reservation_specification = {}
        if capacity_reservation_preference:
            capacity_reservation_specification[
                "CapacityReservationPreference"
            ] = capacity_reservation_preference

        network_interface_attachment_specification = []
        if subnet_id:
            # Subnet id and network interfaces cannot be specified together
            # The network interfaces will be handled in the update process
            network_interface_attachment_specification = None
        elif network_interfaces:
            for network_interface_attachment in network_interfaces:
                network_interface_attachment_specification.append(
                    {
                        "NetworkInterfaceId": network_interface_attachment.get(
                            "network_interface_id"
                        ),
                        "DeviceIndex": network_interface_attachment.get("device_index"),
                        "NetworkCardIndex": network_interface_attachment.get(
                            "network_card_index"
                        ),
                    }
                )

        new_block_device_mappings = None
        if block_device_mappings:
            new_block_device_mappings = list()
            for bd in block_device_mappings:
                device_name = bd.get("device_name")
                if not device_name:
                    continue
                # Create an empty mapping while the instance is starting -- it will be connected on update
                new_bd = dict(DeviceName=device_name, NoDevice="")
                if bd.get("virtual_name"):
                    new_bd["VirtualName"] = bd["virtual_name"]
                new_block_device_mappings.append(new_bd)

        # Create a brand new instance with minimal arguments, it will be updated after creation
        create_ret = await hub.exec.boto3.client.ec2.run_instances(
            ctx,
            ClientToken=client_token,
            MaxCount=1,
            MinCount=1,
            ImageId=image_id,
            InstanceType=instance_type,
            EbsOptimized=ebs_optimized,
            BlockDeviceMappings=new_block_device_mappings,
            KernelId=kernel_id,
            Placement=placement or None,
            SubnetId=subnet_id,
            NetworkInterfaces=network_interface_attachment_specification,
            Monitoring=monitoring_options or None,
            PrivateIpAddress=private_ip_address,
            UserData=user_data,
            DisableApiTermination=disable_api_termination,
            InstanceInitiatedShutdownBehavior=instance_initiated_shutdown_behavior,
            EnclaveOptions=enclave_options or None,
            RamDiskId=ram_disk_id,
            TagSpecifications=tag_specifications,
            ElasticGpuSpecification=elastic_gpu_specifications,
            ElasticInferenceAccelerators=elastic_inference_accelerators,
            IamInstanceProfile={"Arn": iam_instance_profile_arn}
            if iam_instance_profile_arn
            else None,
            KeyName=key_name,
            InstanceMarketOptions=instance_market_options,
            CreditSpecification=credit_specifications or None,
            CpuOptions=cpu_options or None,
            CapacityReservationSpecification=capacity_reservation_specification or None,
            LicenseSpecifications=[
                {"LicenseConfigurationArn": license_arn}
                for license_arn in license_arns or []
            ],
            HibernationOptions=hibernation_options or None,
            MetadataOptions=metadata_options or None,
            PrivateDnsNameOptions=private_dns_name_options or None,
            MaintenanceOptions=maintenance_options or None,
            SecurityGroupIds=security_group_ids,
        )

        result["result"] &= create_ret.result
        if not create_ret:
            result["comment"] += [create_ret.comment]
            return result

        result["comment"] += [f"Created '{name}'"]
        resource_id = create_ret.ret["Instances"][0]["InstanceId"]
        # This makes sure the created Instance is saved to esm regardless if the subsequent update call fails or not.
        result["new_state"] = dict(name=name, resource_id=resource_id)
        result["force_save"] = True

        present_ret = (
            await hub.tool.aws.ec2.instance.state.convert_instance_to_present_async(
                ctx, {"Reservations": [create_ret.ret]}, name=name
            )
        )
        current_state = present_ret[resource_id]

    if not current_state:
        result["comment"] += [
            f"Unable to get the current_state, instance may still be undergoing creation: '{name}'"
        ]
        result["result"] = False
        return result

    # Modify all the attributes that need modification
    if ctx.test:
        result["new_state"] = hub.tool.aws.ec2.instance.state.test(**desired_state)
        return result

    resource = await hub.tool.boto3.resource.create(
        ctx, "ec2", "Instance", id=resource_id
    )

    # TODO: Remove waiters across the board in this flow
    if resource.state["Name"] == "pending":
        await hub.tool.boto3.resource.exec(resource, "wait_until_running")
    elif resource.state["Name"] == "stopping":
        await hub.tool.boto3.resource.exec(resource, "wait_until_stopped")

    try:
        result["result"] &= await hub.tool.aws.ec2.instance.update.init.apply(
            ctx,
            resource,
            old_value=current_state,
            new_value=desired_state,
            comments=result["comment"],
        )
    except Exception as apply_config_error:
        err_msg = f"Failed to apply configuration on the resource {resource_id}: {apply_config_error.__class__.__name__}"
        hub.log.error(err_msg)
        result["comment"].append(err_msg)
        result["result"] = False

    get = await hub.exec.aws.ec2.instance.get(ctx, resource_id=resource_id, name=name)
    if get.result:
        current_state = get.ret
    result["new_state"] = current_state

    if running:
        # If running is set to True then it goes to reconciler and makes sure current state is running
        hub.log.debug(f"Waiting for instance '{resource_id}' to be running")
        result["rerun_data"] = dict(target_instance_state="running")
    else:
        hub.log.debug(f"Waiting for instance '{resource_id}' to be stopped")
        result["rerun_data"] = dict(target_instance_state="stopped")

    return result


async def absent(
    hub, ctx, name: str, resource_id: str = None, client_token: str = None, **kwargs
) -> Dict[str, Any]:
    """Shuts down the specified instance.

    Terminated instances remain visible after termination (for approximately one hour).

    Args:
        name(str): The name of the state.
        resource_id(str): An instance id.
        client_token(str): An idempotence token that was used with the instance creation

    Request Syntax:
        .. code-block:: sls

            [instance-resource-id]:
              aws.ec2.instance.absent:
                - name: "string"
                - resource_id: "string"

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            resource_is_absent:
              aws.ec2.instance.absent:
                - name: value
    """
    result = dict(
        comment=[],
        old_state=ctx.old_state,
        new_state=None,
        name=name,
        result=True,
        rerun_data=dict(target_instance_state="terminated"),
    )

    # Get the resource_id from ESM
    if not resource_id:
        resource_id = (ctx.old_state or {}).get("resource_id")

    if not resource_id and client_token:
        get = await hub.exec.aws.ec2.instance.get(
            ctx, name=name, filters=[{"Name": "client-token", "Values": [client_token]}]
        )
        resource_id = get.ret.resource_id

    # If there still is no resource_id, the instance is gone
    if not resource_id:
        result["comment"] += [f"'{name}' already terminated"]
        return result

    if ctx.test:
        result["comment"] += [f"Would terminate aws.ec2.instance '{name}'"]
        return result

    get = await hub.exec.aws.ec2.instance.get(ctx, name=name, resource_id=resource_id)
    if get.get("ret") is not None:
        result["old_state"] = get["ret"]
        current_status = result["old_state"].get("instance_state")
        if current_status == "terminated":
            result["comment"] += [
                f"Terminated instance '{name}', it will still be visible for about 60 minutes"
            ]
            return result

    ret = await hub.exec.boto3.client.ec2.terminate_instances(
        ctx, InstanceIds=[resource_id]
    )

    if ret.comment:
        result["comment"].append(ret.comment)

    if not ret.result and any("does not exist" in c for c in ret.comment):
        # The command failed because the instance no longer exists
        return result
    elif not ret.result:
        # The failure is some other reason we need to examine
        result["result"] &= ret.result
    else:
        get = await hub.exec.aws.ec2.instance.get(
            ctx, resource_id=resource_id, name=name
        )
        result["new_state"] = get.ret
        result["comment"] += [
            f"Terminated instance '{name}', it will still be visible for about 60 minutes"
        ]

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: bash

            $ idem describe aws.ec2.instance
    """
    result = {}
    ret = await hub.exec.boto3.client.ec2.describe_instances(ctx)

    if not ret:
        hub.log.debug(f"Could not describe Instances: {ret.comment}")
        return {}

    instances = await hub.tool.aws.ec2.instance.state.convert_instance_to_present_async(
        ctx, ret.ret
    )

    for instance_id, present_state in instances.items():
        result[instance_id] = {
            "aws.ec2.instance.present": [{k: v} for k, v in present_state.items()]
        }

    return result


def is_pending(hub, ret):
    new_state = ret.get("new_state", None)
    rerun_data = ret.get("rerun_data", None)
    if new_state and rerun_data:
        current_status = new_state.get("instance_state")
        resource_id = new_state.get("resource_id")
        if current_status and isinstance(current_status, str):
            if current_status.casefold() == rerun_data.get("target_instance_state"):
                hub.log.debug(
                    f"No need to reconcile new state for instance '{resource_id}' with current status {current_status}"
                )
                # Cleanup new_state to make sure instance is removed from ESM
                if (
                    rerun_data.get("target_instance_state") == "terminated"
                    and ret["new_state"]
                ):
                    return True
                return False
            else:
                hub.log.debug(
                    f"Reconcile new state for instance '{resource_id}' with current status {current_status}"
                )
                return True
    if (
        not new_state
        and rerun_data
        and rerun_data.get("target_instance_state") == "terminated"
    ):
        # No need to reconcile terminated state
        return False

    return (not ret["result"]) or bool(ret["changes"])



from collections import OrderedDict
from typing import Any
from typing import Dict

"""
Util functions to convert raw resource state from AWS Config rule to present input format.
"""


async def convert_raw_config_rule_to_present_async(
    hub, ctx, raw_resource: Dict[str, Any], idem_resource_name: str
) -> Dict[str, Any]:
    # ConfigRuleName is the unique identifier for Config so it is set as resource_id
    resource_id = raw_resource.get("ConfigRuleName")
    config_rule_arn = raw_resource.get("ConfigRuleArn")
    tag_ret = await hub.exec.boto3.client.config.list_tags_for_resource(
        ctx, ResourceArn=config_rule_arn
    )

    resource_parameters = OrderedDict(
        {
            "ConfigRuleName": "config_rule_name",
            "ConfigRuleArn": "config_rule_arn",
            "ConfigRuleId": "config_rule_id",
            "Scope": "scope",
            "Source": "source",
            "ConfigRuleState": "config_rule_state",
            "MaximumExecutionFrequency": "max_execution_frequency",
            "InputParameters": "input_parameters",
        }
    )
    resource_translated = {"name": idem_resource_name, "resource_id": resource_id}
    if tag_ret:
        resource_translated["tags"] = hub.tool.aws.tag_utils.convert_tag_list_to_dict(
            tag_ret["ret"]["Tags"]
        )
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource:
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)
    return resource_translated


def convert_raw_config_recorder_to_present(
    hub, ctx, raw_resource: Dict[str, Any]
) -> Dict[str, Any]:
    resource_translated = {
        "name": raw_resource.get("name"),
        "resource_id": raw_resource.get("name"),
        "role_arn": raw_resource.get("roleARN"),
        "recording_group": raw_resource.get("recordingGroup"),
    }
    return resource_translated


def convert_raw_config_delivery_channel_to_present(
    hub, ctx, raw_resource: Dict[str, Any], idem_resource_name: str
) -> Dict[str, Any]:
    # name is the unique identifier for delivery channel so it is set as resource_id
    resource_id = raw_resource.get("name")

    resource_parameters = OrderedDict(
        {
            "name": "name",
            "s3BucketName": "s3_bucket_name",
            "s3KeyPrefix": "s3_key_prefix",
            "s3KmsKeyArn": "s3_kms_key_arn",
            "snsTopicARN": "sns_topic_arn",
            "configSnapshotDeliveryProperties": "config_snapshot_delivery_properties",
        }
    )
    resource_translated = {"name": idem_resource_name, "resource_id": resource_id}
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource:
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)

    if raw_resource.get("configSnapshotDeliveryProperties"):
        config_snapshot_delivery = {}
        config_snapshot_delivery["delivery_frequency"] = raw_resource.get(
            "configSnapshotDeliveryProperties"
        ).get("deliveryFrequency")
        resource_translated[
            "config_snapshot_delivery_properties"
        ] = config_snapshot_delivery

    return resource_translated


def convert_raw_config_recorder_status_to_present(
    hub, ctx, raw_resource: Dict[str, Any], recording: bool = False
) -> Dict[str, Any]:
    resource_translated = {
        "name": raw_resource.get("name"),
        "resource_id": raw_resource.get("name"),
        "recording": raw_resource.get("recording", recording),
    }
    return resource_translated


def convert_raw_config_aggregator_to_present(
    hub, ctx, raw_resource: Dict[str, Any]
) -> Dict[str, Any]:
    """Convert a raw Configuration Aggregator resources to object.

    Args:
        raw_resource(Dict[str, Any] = Optional):
            Raw Configuration Aggregators
    """
    resource_id = raw_resource.get("ConfigurationAggregatorName")

    resource_parameters = OrderedDict(
        {
            "AccountAggregationSources": "account_aggregation_sources",
            "OrganizationAggregationSource": "organization_aggregation_source",
        }
    )
    resource_translated = {"name": resource_id, "resource_id": resource_id}
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource:
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)
    return resource_translated



"""
Helpers for bootstrapping an instance.
These are used for bootstrapping aws ec2 instances, but these functions are not unique to aws.
Eventually they can be moved to a more generic project.
"""
from typing import List


def managers(hub, sub=None, build: str = "") -> List[str]:
    """
    Find all available heist managers.
    """
    result = []

    if sub is None:
        sub = hub.heist

    # Recurse through subs
    for s in sub._subs:
        if not build:
            new_build = s
        else:
            new_build = ".".join((build, s))
        result.extend(hub.tool.heist.bootstrap.managers(sub=sub[s], build=new_build))

    # Loaded mods are the leaf nodes
    for l in sub._loaded:
        if not build:
            new_build = l
        else:
            new_build = ".".join((build, l))
        if "run" in sub[l]._funcs:
            result.append(new_build)

    return result



"""Utility functions for Secretsmanager Tags. Tags related functions for AWS service 'secretsmanager'."""
import copy
from typing import Any
from typing import Dict


async def update_tags(
    hub, ctx, resource_id: str, old_tags: dict, new_tags: dict
) -> Dict[str, Any]:
    r"""
    Updates tags for a given resource.
    """

    result = dict(comment=[], result=True, ret=None)
    tags_to_add = {}
    tags_to_remove = {}
    if new_tags is not None:
        tags_to_remove, tags_to_add = hub.tool.aws.tag_utils.diff_tags_dict(
            old_tags=old_tags, new_tags=new_tags
        )

    if (not tags_to_remove) and (not tags_to_add):
        # If there is nothing to add or remove, return from here with old tags, if present
        result["ret"] = copy.deepcopy(old_tags if old_tags else {})
        return result

    if tags_to_remove:
        if not ctx.get("test", False):
            delete_ret = await hub.exec.boto3.client.secretsmanager.untag_resource(
                ctx, SecretId=resource_id, TagKeys=list(tags_to_remove.keys())
            )
            if not delete_ret["result"]:
                result["comment"] = delete_ret["comment"]
                result["result"] = False
                return result

    if tags_to_add:
        if not ctx.get("test", False):
            add_ret = await hub.exec.boto3.client.secretsmanager.tag_resource(
                ctx,
                SecretId=resource_id,
                Tags=hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags=tags_to_add),
            )
            if not add_ret["result"]:
                result["comment"] += add_ret["comment"]
                result["result"] = False
                return result

    result["ret"] = new_tags
    if ctx.get("test", False):
        result["comment"] = hub.tool.aws.comment_utils.would_update_tags_comment(
            tags_to_remove=tags_to_remove, tags_to_add=tags_to_add
        )
    else:
        result["comment"] = hub.tool.aws.comment_utils.update_tags_comment(
            tags_to_remove=tags_to_remove, tags_to_add=tags_to_add
        )
    return result



"""States module for managing EC2 VPC Endpoint Service Permissions."""
import copy
from typing import Any
from typing import Dict

from dict_tools import differ

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    service_id: str,
    principal_arn: str,
    resource_id: str = None,
) -> Dict[str, Any]:
    """
    Modifies the permissions for your VPC endpoint service. You can add or remove permissions for service consumers
    (Amazon Web Services accounts, users, and IAM roles) to connect to your endpoint service.

    If you grant permissions to all principals, the service is public. Any users who know the name of a public service can send a
    request to attach an endpoint. If the service does not require manual approval, attachments are automatically
    approved.

    Args:
        name(str): Idem name of the resource.

        service_id(str): The ID of the service.

        principal_arn(str, Optional): The Amazon Resource Name (ARN) of the principal. To grant permissions to all principals, specify an asterisk (*).

        resource_id(str): The ID of the service permission. Defaults to None.

    Returns:
        Dict[str, Any]

    Example:
        .. code-block:: sls

          my-vpc-endpoint-service-permission:
            aws.ec2.vpc_endpoint_service_permission.present:
              - service_id: value
              - principal_arn: value
    """
    result = dict(
        comment=[], old_state={}, new_state={}, name=name, result=True, rerun_data=None
    )

    desired_state = {
        k: v
        for k, v in locals().items()
        if k not in ("hub", "ctx", "kwargs", "result") and v is not None
    }

    current_state = None

    if resource_id:
        before = await hub.exec.aws.ec2.vpc_endpoint_service_permission.get(
            ctx, name=name, service_id=service_id, principal_arn=principal_arn
        )

        if not before["result"] or not before["ret"]:
            result["result"] = False
            result["comment"] = before["comment"]
            return result

        result["old_state"] = current_state = copy.deepcopy(before["ret"])
        result["comment"].append(
            f"'aws.ec2.vpc_endpoint_service_permission: {name}' already exists"
        )

    if current_state:
        # If provided principal is not in current set of principals, add it
        needs_update = (
            "principal_arn" in current_state
            and current_state["principal_arn"] != principal_arn
        )

        if needs_update:
            if ctx.test:
                result["new_state"] = hub.tool.test_state_utils.generate_test_state(
                    enforced_state={}, desired_state=desired_state
                )
                result["comment"].append(
                    f"Would update aws.ec2.vpc_endpoint_service_permission '{name}'"
                )
                return result
            else:
                update_ret = await hub.exec.aws.ec2.vpc_endpoint_service_permission.update(
                    ctx,
                    **{
                        "service_id": service_id,
                        # requested becomes new principal to add
                        "add_allowed_principals": [principal_arn],
                        # current becomes principal to be removed
                        "remove_allowed_principals": [current_state["principal_arn"]],
                    },
                )
                result["result"] = update_ret["result"]

                if result["result"]:
                    result["comment"].append(
                        f"Updated aws.ec2.vpc_endpoint_service_permission '{name}'"
                    )
                else:
                    result["comment"].append(update_ret["comment"])
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={}, desired_state=desired_state
            )
            result["comment"].append(
                f"Would create aws.ec2.vpc_endpoint_service_permission '{name}'"
            )
            return result

        create_ret = await hub.exec.aws.ec2.vpc_endpoint_service_permission.create(
            ctx, **{"service_id": service_id, "add_allowed_principals": [principal_arn]}
        )
        result["result"] = create_ret["result"]

        if result["result"]:
            result["comment"] = hub.tool.aws.comment_utils.create_comment(
                resource_type="aws.ec2.vpc_endpoint_service_permission", name=name
            )
            resource_id = create_ret["ret"].get("resource_id")
            # Safeguard for any future errors so that the resource_id is saved in the ESM
            result["new_state"] = dict(name=name, resource_id=resource_id)
        else:
            result["comment"].append(create_ret["comment"])

    if not result["result"]:
        # If there is any failure in create/update, it should reconcile.
        # The type of data is less important here to use default reconciliation
        # If there are no changes for 3 runs with rerun_data, then it will come out of execution
        result["rerun_data"] = dict(name=name, resource_id=resource_id)

    after = await hub.exec.aws.ec2.vpc_endpoint_service_permission.get(
        ctx, name=name, service_id=service_id, principal_arn=principal_arn
    )
    result["new_state"] = after.ret
    return result


async def absent(
    hub,
    ctx,
    name: str,
    service_id: str,
    principal_arn: str,
    resource_id: str = None,
) -> Dict[str, Any]:
    """
    Removes the permissions for your VPC endpoint service. You can remove permissions for service consumers
    (Amazon Web Services accounts, users, and IAM roles) to connect to your endpoint service.

    Args:
        name(str): Idem name of the resource.

        service_id(str): The ID of the service.

        principal_arn(str, Optional): The Amazon Resource Name (ARN) of the principal.

        resource_id(str, Optional): The ID of the service permission.

    Returns:
        Dict[str, Any]

    Example:
        .. code-block:: sls

            my-vpc-endpoint-service-permission:
              aws.ec2.vpc_endpoint_service_permission.absent:
                - service_id: value
                - principal_arn: value
    """

    result = dict(
        comment=[], old_state={}, new_state={}, name=name, result=True, rerun_data=None
    )

    if not resource_id:
        resource_id = (ctx.old_state or {}).get("resource_id")

    # This is to make absent idempotent. If absent is run again, it would be a no-op
    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.ec2.vpc_endpoint_service_permission", name=name
        )
        return result

    before = await hub.exec.aws.ec2.vpc_endpoint_service_permission.get(
        ctx, name=name, service_id=service_id, principal_arn=principal_arn
    )

    # Case: Error
    if not before["result"]:
        result["result"] = False
        result["comment"] = before["comment"]
        return result

    # Case: Not Found
    if not before["ret"]:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.ec2.vpc_endpoint_service_permission", name=name
        )
        return result

    result["old_state"] = before["ret"]

    if ctx.get("test", False):
        result["comment"].append(
            f"Would delete aws.ec2.vpc_endpoint_service_permission '{name}'",
        )
        return result

    delete_ret = await hub.exec.aws.ec2.vpc_endpoint_service_permission.delete(
        ctx, name=name, service_id=service_id, remove_allowed_principals=[principal_arn]
    )

    result["result"] = delete_ret["result"]

    if not result["result"]:
        # If there is any failure in delete, it should reconcile.
        # The type of data is less important here to use default reconciliation
        # If there are no changes for 3 runs with rerun_data, then it will come out of execution
        result["rerun_data"] = resource_id
        result["comment"].append(delete_ret["result"])

    result["comment"] = hub.tool.aws.comment_utils.delete_comment(
        resource_type="aws.ec2.vpc_endpoint_service_permission", name=name
    )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """
    Describes the principals (service consumers) that are permitted to discover your VPC endpoint service.

    Returns:
        Dict[str, Any]

    Example:

        .. code-block:: bash

            $ idem describe aws.ec2.vpc_endpoint_service_permission
    """

    result = {}

    services_ret = await hub.exec.aws.ec2.vpc_endpoint_service_configuration.list(ctx)

    if not services_ret or not services_ret["result"]:
        hub.log.debug(
            f"Could not describe aws.ec2.vpc_endpoint_service_configuration {services_ret['comment']}"
        )
        return result

    for service_resource in services_ret["ret"]:
        service_id = service_resource.get("resource_id")

        ret = await hub.exec.aws.ec2.vpc_endpoint_service_permission.list(
            ctx, service_id
        )

        if not ret or not ret["result"]:
            hub.log.debug(
                f"Could not describe aws.ec2.vpc_endpoint_service_permission {ret['comment']}"
            )

        for service_permission_resource in ret["ret"]:
            resource_id = service_permission_resource.get("resource_id")
            result[resource_id] = {
                "aws.ec2.vpc_endpoint_service_permission.present": [
                    {parameter_key: parameter_value}
                    for parameter_key, parameter_value in service_permission_resource.items()
                ]
            }

    return result



"""State module for managing AWS WAFV2 IPSet."""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    scope: str,
    ip_address_version: str,
    addresses: list,
    description: str = None,
    resource_id: str = None,
    tags: List[
        make_dataclass(
            """Tags associated with IPSet resources in wafv2""" "Tags",
            [("Key", str), ("Value", str, field(default=None))],
        )
    ] = None,
    lock_token: str = None,
) -> Dict[str, Any]:
    """Creates an AWS IPSet for AWS WAFV2.

    Args:
        name (str): The name of the IP set on AWS. You cannot change the name of an IPSet after you create it.

        scope (str): Specifies whether this is for an Amazon CloudFront distribution or for a regional
             application. A ``REGIONAL`` application can be an Application Load Balancer (ALB), an Amazon API Gateway
             REST API, or an AppSync GraphQL API.
             To work with CloudFront, you must also specify the Region US East (N. Virginia) as follows:
             CLI - Specify the Region when you use the CloudFront scope: ``--scope=CLOUDFRONT --region=us-east-1`` .
             API and SDKs - For all calls, use the Region endpoint us-east-1.

        description (str, Optional): A description of the IPSet that helps with identification.

        ip_address_version (str): The version of the IP addresses, either IPV4 or IPV6 .

        addresses (list): Contains an array of strings that specifies zero or more IP addresses or blocks of IP addresses in Classless Inter-Domain Routing (CIDR) notation. WAF supports all IPv4 and IPv6 CIDR ranges except for /0.

        resource_id(str, Optional): Needed to retrieve/update/delete description of an existing IPSet. It is a unique identifier for the set. This Id is returned in the responses to create and list commands.

        lock_token (str, Optional): A token used for optimistic locking. WAF returns a token to your get and list requests, to mark the state of the entity at the time of the request. To make changes to the entity associated with the token, you provide the token to operations like update and delete. WAF uses the token to ensure that no changes have been made to the entity since you last retrieved it. If a change has been made, the update fails with a WAFOptimisticLockException . If this happens, perform another get , and use the new token returned by that operation. Example address strings: To configure WAF to allow, block, or count requests that originated from the IP address 192.12.2.44, specify 192.12.2.44/32. To configure WAF to allow, block, or count requests that originated from IP addresses from 193.0.2.1 to 192.0.2.255, specify 192.0.2.0/24. To configure WAF to allow, block, or count requests that originated from the IP address 1111:0000:0000:0000:0000:0000:0000:0111, specify 1111:0000:0000:0000:0000:0000:0000:0111/128. To configure WAF to allow, block, or count requests that originated from IP addresses 1111:0000:0000:0000:0000:0000:0000:0000 to 1111:0000:0000:0000:ffff:ffff:ffff:ffff, specify 1111:0000:0000:0000:0000:0000:0000:0000/64. For more information about CIDR notation, see the Wikipedia entry Classless Inter-Domain Routing. Example JSON Addresses specifications: Empty array: "Addresses": [] Array with one address: "Addresses": ["192.0.2.44/32"] Array with three addresses: "Addresses": ["192.0.2.44/32", "192.0.2.0/24", "192.0.0.0/16"] INVALID specification: "Addresses": [""] INVALID(str) --

        tags (Dict or List, Optional): An array of key:value pairs to associate with the resource.

             Key (str): Part of the key:value pair that defines a tag. You can use a tag key to
             describe a category of information, such as ``"customer."`` Tag keys are case-sensitive.

             Value (str): Part of the key:value pair that defines a tag. You can use a tag value
             to describe a specific value within a category, such as "companyA" or "companyB." Tag values are case-sensitive.

    Request Syntax:
        .. code-block:: sls

            [my_idem_ip_set_present]:
                 aws.wafv2.ip_set.present:
                     - name: 'str'
                     - scope: 'str'
                     - ip_address_version: 'str'
                     - addresses: 'str'
                         - 192.0.1.0/24
                     - resource_id: 'str'
                     - description: 'str'
                     - tags:
                         - Key: 'str'
                         Value: 'str'


    Returns:
        Dict[str, Any]

    Example:
        .. code-block:: sls

         [my_idem_ip_set_present]:
             aws.wafv2.ip_set.present:
                 - name: test-wafv2-ip-set-1663200539
                 - scope: REGIONAL
                 - ip_address_version: IPV4
                 - addresses:
                     - 192.0.2.44/32
                     - 19.1.2.44/32
                 - resource_id: f547799f-e325-4876-9e0e-264d835eb3fd
                 - description: for testing idem plugin for IPSet.
                 - tags:
                     - Key: Name
                       Value: ipset-name
                     - Key: test-wafv2-ip-set-1663200539
                       Value: test-wafv2-ip-set-1663200539
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    resource_updated: bool = False
    if resource_id:
        # IPSet exists , update
        before_ret = await hub.exec.aws.wafv2.ip_set.get(
            ctx, name=name, scope=scope, resource_id=resource_id
        )
        if not before_ret["result"] or not before_ret["ret"]:
            result["result"] = False
            result["comment"] = before_ret["comment"]
            return result

        result["old_state"] = copy.deepcopy(before_ret["ret"])
        result["new_state"] = copy.deepcopy(result["old_state"])
        arn = before_ret["ret"]["arn"]
        old_tags = before_ret["ret"]["tags"]
        desired_state = {
            "name": name,
            "scope": scope,
            "resource_id": resource_id,
            "description": description,
            "addresses": addresses,
        }
        # Update IPSet
        update_ret = await hub.tool.aws.wafv2.ip_set_utils.update(
            ctx, name=name, current_state=before_ret["ret"], desired_state=desired_state
        )
        result["comment"] = update_ret["comment"]
        if not update_ret["result"]:
            result["result"] = False
            return result
        result["result"] = update_ret["result"]

        resource_updated = resource_updated or bool(update_ret["ret"])
        if update_ret["ret"] and ctx.get("test", False):
            result["new_state"].update(copy.deepcopy(update_ret["ret"]))

        # Update Tags
        if tags is not None and tags != result["old_state"].get("tags"):
            update_tags_ret = await hub.tool.aws.wafv2.tag.update_tags(
                ctx,
                resource_arn=arn,
                old_tags=old_tags,
                new_tags=tags,
            )
            result["comment"] = result["comment"] + update_tags_ret["comment"]
            result["result"] = result["result"] and update_tags_ret["result"]
            if not update_tags_ret["result"]:
                result["result"] = False
                hub.log.debug(f"Failed updating tags for aws.wafv2.ip.set '{name}'")
                return result
            resource_updated = resource_updated or bool(update_tags_ret["ret"])
            if update_tags_ret["ret"] and ctx.get("test", False):
                result["new_state"]["tags"] = update_tags_ret["ret"]

        if ctx.get("test", False) and resource_updated:
            return result

    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "scope": scope,
                    "description": description,
                    "ip_address_version": ip_address_version,
                    "addresses": addresses,
                    "tags": tags,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.wafv2.ip_set", name=name
            )
            return result
        else:
            # IPSet not present , create
            create_ret = await hub.exec.boto3.client.wafv2.create_ip_set(
                ctx,
                Name=name,
                Scope=scope,
                Description=description,
                IPAddressVersion=ip_address_version,
                Addresses=addresses,
                Tags=tags
                if type(tags) == list
                else hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags),
            )
            if not create_ret["result"]:
                result["result"] = False
                result["comment"] = create_ret["comment"]
                return result
            resource_id = create_ret["ret"]["Summary"]["Id"]
            result["comment"] = hub.tool.aws.comment_utils.create_comment(
                resource_type="aws.wafv2.ip_set", name=name
            )

    if (not result["old_state"]) or resource_updated:
        after_ret = await hub.exec.aws.wafv2.ip_set.get(
            ctx, name=name, scope=scope, resource_id=resource_id
        )
        if not after_ret["result"]:
            result["result"] = False
            result["comment"] += after_ret["comment"]
            return result
        result["new_state"] = after_ret["ret"]
    else:
        result["new_state"] = copy.deepcopy(result["old_state"])
    return result


async def absent(
    hub, ctx, name: str, scope: str, resource_id: str = None
) -> Dict[str, Any]:
    """Deletes the IPSet for the AWS WAFV2 associated with a Web ACL.

    Args:
        name(str):
            The name of the IP set. You cannot change the name of an IPSet after you create it.
        scope(str):
            Specifies whether this is for an Amazon CloudFront distribution or for a Regional application.
            A ``Regional`` application can be an Application Load Balancer (ALB), an Amazon API Gateway REST API, or an AppSync GraphQL API.
            To work with CloudFront, you must also specify the Region US East (N. Virginia) as follows:
            CLI - Specify the Region when you use the CloudFront scope: ``--scope=CLOUDFRONT --region=us-east-1`` .
            API and SDKs - For all calls, use the Region endpoint us-east-1.
            For regional applications, you can use any of the endpoints in the list. A regional application can be an
            Application Load Balancer (ALB), an Amazon API Gateway REST API, an AppSync GraphQL API, or an Amazon Cognito user pool.
            For Amazon CloudFront applications, you must use the API endpoint listed for US East (N. Virginia): us-east-1.
        resource_id(str, Optional):
            A unique identifier for the set. This ID is returned in the responses to create and list commands.
            You provide it to operations like update and delete. Idem automatically considers this resource being absent
            if this field is not specified.

    Request Syntax:
        .. code-block:: sls

            [my_idem_ip_set_absent_test]:
                aws.wafv2.ip_set.absent:
                    - name: 'string'
                    - scope: 'string'
                    - resource_id: 'string'

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            [my_idem_ip_set_absent_test]:
                aws.wafv2.ip_set.absent:
                    - name: test-wafv2-ip-set-1663200539
                    - scope: REGIONAL
                    - resource_id: f547799f-e325-4876-9e0e-264d835eb3fd
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.wafv2.ip_set", name=name
        )
        return result
    before = await hub.exec.aws.wafv2.ip_set.get(
        ctx,
        name=name,
        scope=scope,
        resource_id=resource_id,
    )
    if not before["result"]:
        result["comment"] = before["comment"]
        result["result"] = False
        return result
    if not before["ret"]:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.wafv2.ip_set", name=name
        )
    elif ctx.get("test", False):
        result["old_state"] = before["ret"]
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.wafv2.ip_set", name=name
        )
        return result
    else:
        lock_token = before["ret"]["lock_token"]
        result["old_state"] = before["ret"]
        delete_ret = await hub.exec.boto3.client.wafv2.delete_ip_set(
            ctx, Name=name, Scope=scope, Id=resource_id, LockToken=lock_token
        )
        result["result"] = delete_ret["result"]
        if not result["result"]:
            result["comment"] = delete_ret["comment"]
            result["result"] = False
            return result
        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.wafv2.ip_set", name=name
        )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """The goal of the describe function is to provide users a way to generate an SLS output.

    This sls output can be used to regenerate or update the IPSet resource on AWS.

    Returns:
        Dict[str, Dict[str, Any]]

    Examples:
        Calling from the CLI:
            .. code-block:: bash

                idem describe aws.wafv2.ip_set --acct-profile <environment>
    """
    result = {}
    scopes = ["CLOUDFRONT", "REGIONAL"]
    for scope in scopes:
        ret = await hub.exec.aws.wafv2.ip_set.list(ctx, scope=scope)
        if not ret["result"]:
            hub.log.debug(f"Could not list ip_set " f"{ret['comment']}")
            continue
        for ip_set in ret["ret"]:
            ipset_name = ip_set["ret"]["name"]
            result[ipset_name] = {
                "aws.wafv2.ip_set.present": [
                    {parameter_key: parameter_value}
                    for parameter_key, parameter_value in ip_set["ret"].items()
                ]
            }
    return result



"""Exec module to get AWS ELB route53 Hosted Zone ID based on the region specified"""
from typing import Any
from typing import Dict

AWS_REGION_TO_HOSTED_ZONE_MAP = {
    "af-south-1": "Z268VQBMOI5EKX",
    "ap-east-1": "Z3DQVH9N71FHZ0",
    "ap-northeast-1": "Z14GRHDCWA56QT",
    "ap-northeast-2": "ZWKZPGTI48KDX",
    "ap-northeast-3": "Z5LXEXXYW11ES",
    "ap-south-1": "ZP97RAFLXTNZK",
    "ap-southeast-1": "Z1LMS91P8CMLE5",
    "ap-southeast-2": "Z1GM3OXH4ZPM65",
    "ap-southeast-3": "Z08888821HLRG5A9ZRTER",
    "ca-central-1": "ZQSVJUPU6J1EY",
    "cn-north-1": "Z1GDH35T77C1KE",
    "cn-northwest-1": "ZM7IZAIOVVDZF",
    "eu-central-1": "Z215JYRZR1TBD5",
    "eu-north-1": "Z23TAZ6LKFMNIO",
    "eu-south-1": "Z3ULH7SSC9OV64",
    "eu-west-1": "Z32O12XQLNTSW2",
    "eu-west-2": "ZHURV8PSTC4K8",
    "eu-west-3": "Z3Q77PNBQS71R4",
    "me-south-1": "ZS929ML54UICD",
    "sa-east-1": "Z2P70J7HTTTPLU",
    "us-east-1": "Z35SXDOTRQ7X7K",
    "us-east-2": "Z3AADJGX6KTTL2",
    "us-gov-east-1": "Z166TLBEWOO7G0",
    "us-gov-west-1": "Z33AYJ8TM3BH4J",
    "us-west-1": "Z368ELLRRE2KJ0",
    "us-west-2": "Z1H1FL5HABSF5",
    "eu-south-2": "Z0956581394HF5D5LXGAP",
    "eu-central-2": "Z06391101F2ZOEP8P5EB3",
    "ap-south-2": "Z0173938T07WNTVAEPZN",
    "me-central-1": "Z08230872XQRWHG2XF6I",
}


async def get(hub, ctx, name, region: str = None) -> Dict[str, Any]:
    """
    Returns the AWS ELB route53 Hosted Zone ID of the current region whose credentials are used to call the operation.
    Refer-: https://docs.aws.amazon.com/general/latest/gr/elb.html

    Args:
        name(str): An Idem name of the resource.
        region(str, Optional): AWS region for which ELB route53 Hosted Zone ID is required.
            If not specified region specified in credentials file will be used.

    Request Syntax:
        [Idem-resource-state-name]:
          exec.run:
          - path: aws.elb.elb_hosted_zone_id.get
          - kwargs:
              region: 'string'

    Examples:

        my-elb-hosted_zone-id:
          exec.run:
            - path: aws.elb.elb_hosted_zone_id.get
            - kwargs:
                region: us-west-2

    Response Syntax:
          {
            'elb_hosted_zone_id': 'string',
            'name': 'string'
          }

    Response Structure:
        name(str): An Idem name of the resource.
        elb_hosted_zone_id(str): AWS ELB route53 Hosted Zone ID of the AWS region used in credentials

    """
    result = dict(comment=(), ret=None, result=True)
    region = region or ctx["acct"].get("region_name")
    region_elb_hosted_zone_id = AWS_REGION_TO_HOSTED_ZONE_MAP.get(region)
    if not region_elb_hosted_zone_id:
        result["result"] = False
        result["comment"] = (
            f"AWS ELB route53 Hosted Zone ID not found for region {region}",
        )
        return result
    resource_translated = {
        "name": name,
        "elb_hosted_zone_id": region_elb_hosted_zone_id,
    }

    result["ret"] = resource_translated
    return result



"""Exec module for managing Amazon Elasticsearch Service Domain"""


async def get(
    hub,
    ctx,
    name: str,
    resource_id: str,
):
    """Get the domain configuration information about the specified Elasticsearch domain

    Args:
        name(str): An Idem name of the resource.
        resource_id(str): The name of the Elasticsearch domain.

    Returns:
        .. code-block:: python

           {"result": True|False, "comment": A message List, "ret": None|Dict}

    Examples:
        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: aws.es.elasticsearch_domain.get
                - kwargs:
                    name: my_resource
                    resource_id: resource_id

        Calling this exec function from the cli with resource_id:

        .. code-block:: bash

            idem exec aws.es.elasticsearch_domain.get resource_id="resource_id" name="name"

    """
    result = dict(comment=[], ret=None, result=True)

    get_domain_ret = await hub.exec.boto3.client.es.describe_elasticsearch_domain(
        ctx, DomainName=resource_id
    )

    if not get_domain_ret["result"]:
        if "ResourceNotFoundException" in str(get_domain_ret["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.es.elasticsearch_domain", name=name
                )
            )
            result["comment"] += list(get_domain_ret["comment"])
            return result
        result["result"] = False
        result["comment"] = list(get_domain_ret["comment"])
        return result

    elasticsearch_domain = get_domain_ret["ret"].get("DomainStatus")
    arn = elasticsearch_domain.get("ARN")

    tags_ret = await hub.exec.boto3.client.es.list_tags(ctx, ARN=arn)
    tags = []
    if not tags_ret["result"]:
        result["comment"] = list(tags_ret["comment"])
        result["result"] = False
        return result
    else:
        if tags_ret["ret"] and tags_ret.get("ret").get("TagList"):
            tags = tags_ret.get("ret").get("TagList")

    result[
        "ret"
    ] = hub.tool.aws.es.elasticsearch_domain.convert_raw_elasticsearch_domain_to_present(
        elasticsearch_domain,
        tags=hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags),
    )

    return result



from typing import Dict
from typing import List


async def apply(
    hub,
    ctx,
    resource,
    *,
    old_value: List[Dict[str, str]],
    new_value: List[Dict[str, str]],
    comments: List[str],
) -> bool:
    """
    Modify an ec2 instance based on a single parameter in it's "present" state

      - block_device_mappings:
        - device_name: /dev/xvda
          volume_id : vol-0f0c6dac5d13b9dc4
          delete_on_termination: False

    Args:
        hub:
        ctx: The ctx from a state module call
        resource: An ec2 instance resource object
        old_value: The previous value from the attributes of an existing instance
        new_value: The desired value from the ec2 instance present state parameters
        comments: A running list of comments abound the update process
    """
    old_volume_attachments = {
        bdm["device_name"]: bdm for bdm in old_value if "device_name" in bdm
    }
    new_volume_attachments = {
        bdm["device_name"]: bdm for bdm in new_value if "device_name" in bdm
    }

    # If a device name is in the old mappings, but not the new, it's a volume to detach
    volumes_to_remove = set(old_volume_attachments.keys()) - set(
        new_volume_attachments.keys()
    )
    # If a device name is in the new mappings, but not the old, its a new volume to attach
    volumes_to_attach = set(new_volume_attachments.keys()) - set(
        old_volume_attachments.keys()
    )

    # If a volume exists in both places then check if the volume_id has changed
    volumes_to_move = set(old_volume_attachments.keys()).intersection(
        set(new_volume_attachments.keys())
    )
    for device_name in volumes_to_move:
        if (
            old_volume_attachments[device_name]["volume_id"]
            != new_volume_attachments[device_name]["volume_id"]
        ):
            # Detach the volume from the old location and attach it to the new location
            volumes_to_remove[device_name] = old_volume_attachments[device_name]
            volumes_to_attach[device_name] = new_volume_attachments[device_name]

    # Allow moving the root device, but not detaching it without putting something else in its place
    if (
        resource.root_device_name in volumes_to_remove
        and resource.root_device_name not in volumes_to_attach
    ):
        volumes_to_remove.remove(resource.root_device_name)

    # Detach old volumes
    for device_name in volumes_to_remove:
        volume_id = old_volume_attachments[device_name]["volume_id"]
        ret = await hub.exec.boto3.client.ec2.detach_volume(
            ctx,
            Device=device_name,
            InstanceId=resource.id,
            VolumeId=volume_id,
            Force=True,
        )
        if ret.comment:
            comments.append(ret.comment)
        if not ret.result:
            comments.append(f"Could not detach volume: {volume_id} from instance")
            return False

    # Attach new volumes
    for device_name in volumes_to_attach:
        volume_id = new_volume_attachments[device_name]["volume_id"]
        ret = await hub.exec.boto3.client.ec2.attach_volume(
            ctx, Device=device_name, InstanceId=resource.id, VolumeId=volume_id
        )
        if ret.comment:
            comments.append(ret.comment)
        if not ret.result:
            comments.append(f"Could not attach volume: {volume_id} to instance")
            return False

    # modify attachment attributes as needed
    block_device_mappings = list()
    for device_name in new_volume_attachments:
        old_bdm = old_volume_attachments.get(device_name)
        new_bdm = new_volume_attachments[device_name]

        if new_bdm != old_bdm:
            new_bdm_attachment = dict(
                DeviceName=new_bdm["device_name"],
                Ebs=dict(
                    DeleteOnTermination=new_bdm.get("delete_on_termination", True),
                    VolumeId=new_bdm["volume_id"],
                ),
            )
            block_device_mappings.append(new_bdm_attachment)

        if block_device_mappings:
            response = resource.modify_attribute(
                BlockDeviceMappings=block_device_mappings
            )
            if response["ResponseMetadata"]["HTTPStatusCode"] != 200:
                comments.append(
                    f"Unable to set delete_on_termination for volume: {device_name}"
                )
                return False
    return True



from typing import Any
from typing import Dict
from typing import List


async def modify_db_parameter_group(
    hub,
    ctx,
    resource_name,
    old_parameters: List[Dict[str, Any]],
    new_parameters: List[Dict[str, Any]],
):
    """
    Modifies the parameters of a DB parameter group.
    Up to 20 parameters can be modified in a single call.

    Args:
        hub:
        ctx:
        resource_name: aws resource name
        old_parameters: list of old parameters, each with ParameterName , ParameterValue , and ApplyMethod
        new_parameters: list of new parameters, each with ParameterName , ParameterValue , and ApplyMethod

    Returns:
        ret contains the full list of parameters after modification
        {"result": True|False, "comment": "A message", "ret": None}

    """
    parameters_to_modify = _get_parameters_to_modify(old_parameters, new_parameters)

    result = dict(comment=(), result=True, ret=None)
    if not parameters_to_modify:
        return result
    else:
        if not ctx.get("test", False):
            update_parameters_ret = (
                await hub.exec.boto3.client.rds.modify_db_parameter_group(
                    ctx,
                    DBParameterGroupName=resource_name,
                    Parameters=parameters_to_modify,
                )
            )
            if not update_parameters_ret["result"]:
                result["comment"] = update_parameters_ret["comment"]
                result["result"] = False
                return result

        result["ret"] = {
            "parameters": _merge_params(old_parameters, parameters_to_modify)
        }
        result["comment"] = result["comment"] + (
            f"Update parameters: Modified {[key.get('ParameterName') for key in parameters_to_modify]}",
        )
    return result


async def modify_db_cluster_parameter_group(
    hub,
    ctx,
    resource_name,
    old_parameters: List[Dict[str, Any]],
    new_parameters: List[Dict[str, Any]],
):
    """
    Modifies the parameters of a DB cluster parameter group
    Up to 20 parameters can be modified in a single call.

    Args:
        hub:
        ctx:
        resource_name: aws resource name
        old_parameters: list of old parameters, each with ParameterName , ParameterValue , and ApplyMethod
        new_parameters: list of new parameters, each with ParameterName , ParameterValue , and ApplyMethod

    Returns:
        ret contains the full list of parameters after modification
        {"result": True|False, "comment": "A message", "ret": None}

    """
    parameters_to_modify = _get_parameters_to_modify(old_parameters, new_parameters)

    result = dict(comment=(), result=True, ret=None)
    if not parameters_to_modify:
        return result
    else:
        if not ctx.get("test", False):
            update_parameters_ret = (
                await hub.exec.boto3.client.rds.modify_db_cluster_parameter_group(
                    ctx,
                    DBClusterParameterGroupName=resource_name,
                    Parameters=parameters_to_modify,
                )
            )
            if not update_parameters_ret["result"]:
                result["comment"] = update_parameters_ret["comment"]
                result["result"] = False
                return result

        result["ret"] = {
            "parameters": _merge_params(old_parameters, parameters_to_modify)
        }
        result["comment"] = result["comment"] + (
            f"Update parameters: Modified {[key.get('ParameterName') for key in parameters_to_modify]}",
        )
    return result


def _get_parameters_to_modify(
    old_parameters: List[Dict[str, Any]], new_parameters: List[Dict[str, Any]]
):
    parameters_to_modify = []
    old_parameters_map = {
        parameter.get("ParameterName"): parameter for parameter in old_parameters or []
    }

    if new_parameters is not None:
        for parameter in new_parameters:
            if parameter.get("ParameterName") in old_parameters_map:
                name = parameter.get("ParameterName")
                if parameter.get("ParameterValue") != old_parameters_map.get(name).get(
                    "ParameterValue"
                ) or parameter.get("ApplyMethod") != old_parameters_map.get(name).get(
                    "ApplyMethod"
                ):
                    parameters_to_modify.append(parameter)
            else:
                parameters_to_modify.append(parameter)

    return parameters_to_modify


def _merge_params(
    old_parameters: List[Dict[str, Any]], changed_parameters: List[Dict[str, Any]]
):
    # Merge the new parameter value on top of the old values
    old_parameters_map = {
        parameter.get("ParameterName"): parameter for parameter in old_parameters or []
    }

    if changed_parameters:
        for new_param in changed_parameters:
            param_name = new_param.get("ParameterName")
            if param_name in old_parameters_map:
                old_parameters_map[param_name] = {
                    "ParameterName": param_name,
                    "ParameterValue": new_param.get("ParameterValue"),
                    "ApplyMethod": new_param.get("ApplyMethod"),
                }
            else:
                old_parameters_map = {
                    param_name: {
                        "ParameterName": param_name,
                        "ParameterValue": new_param.get("ParameterValue"),
                        "ApplyMethod": new_param.get("ApplyMethod"),
                    }
                }

    return list(old_parameters_map.values())
